"""pytest for the table.py file. Covers the PyTable and PyTables classes.
"""
from test.config import testing_parameters, testingtablename
import pytest
import pandas as pd
from datetime import datetime


@pytest.mark.parametrize("model", testing_parameters)
def test_row_count(model):
    """Tests for `Row_Count()` of PyTable class."""
    assert model.Tables[testingtablename].Row_Count() > 0


@pytest.mark.parametrize("model", testing_parameters)
def test_refresh(model):
    """Tests for `Refresh()` of PyTable class."""
    assert isinstance(model.Tables[testingtablename].Refresh(), pd.DataFrame)


@pytest.mark.parametrize("model", testing_parameters)
def test_last_refresh(model):
    """Tests for `Last_Refresh()` of PyTable class."""
    assert isinstance(model.Tables[testingtablename].Last_Refresh(), datetime)


@pytest.mark.parametrize("model", testing_parameters)
def test_related(model):
    """Tests for `Related()` of PyTable class."""
    assert isinstance(model.Tables[testingtablename].Related(), type(model.Tables))


@pytest.mark.parametrize("model", testing_parameters)
def test_refresh_pytables(model):
    """Tests for `Refresh()` of PyTables class."""
    assert isinstance(model.Tables.Find(testingtablename).Refresh(), pd.DataFrame)


@pytest.mark.parametrize("model", testing_parameters)
def test_query_all_pytables(model):
    """Tests for `Query_All()` of PyTables class."""
    assert isinstance(model.Tables.Query_All(), pd.DataFrame)


@pytest.mark.parametrize("model", testing_parameters)
def test_find_zero_rows_pytables(model):
    """Tests for `Find_Zero_Rows()` of PyTables class."""
    assert isinstance(model.Tables.Find_Zero_Rows(), type(model.Tables))


@pytest.mark.parametrize("model", testing_parameters)
def test_last_refresh_pytables_true(model):
    """Tests group_partition=True for `Last_Refresh()` of PyTables class."""
    assert isinstance(model.Tables.Last_Refresh(group_partition=True), pd.DataFrame)


@pytest.mark.parametrize("model", testing_parameters)
def test_last_refresh_pytables_false(model):
    """Tests group_partition=False for `Last_Refresh()` of PyTables class."""
    assert isinstance(model.Tables.Last_Refresh(group_partition=False), pd.DataFrame)



import pytest
import pytabular as p
from test.config import testing_parameters
from os import getcwd


@pytest.mark.parametrize("model", testing_parameters)
def test_bpa(model):
    te2 = p.Tabular_Editor().EXE
    bpa = p.BPA().Location
    assert isinstance(model.Analyze_BPA(te2, bpa), list)


@pytest.mark.parametrize("model", testing_parameters)
def test_te2_custom_file_path(model):
    assert isinstance(p.Tabular_Editor(getcwd()), p.Tabular_Editor)


@pytest.mark.parametrize("model", testing_parameters)
def test_bpa_custom_file_path(model):
    assert isinstance(p.BPA(getcwd()), p.BPA)



"""pytest for the column.py file. Covers the PyColumn and PyColumns classes.
"""
from test.config import testing_parameters, testingtablename
import pytest
import pandas as pd
from numpy import int64


@pytest.mark.parametrize("model", testing_parameters)
def test_values(model):
    """Tests for `Values()` of PyColumn class."""
    vals = model.Tables[testingtablename].Columns[1].Values()
    assert isinstance(vals, pd.DataFrame) or isinstance(vals, int)


@pytest.mark.parametrize("model", testing_parameters)
def test_distinct_count_no_blank(model):
    """Tests No_Blank=True for `Distinct_Count()` of PyColumn class."""
    vals = model.Tables[testingtablename].Columns[1].Distinct_Count(No_Blank=True)
    assert isinstance(vals, int64)


@pytest.mark.parametrize("model", testing_parameters)
def test_distinct_count_blank(model):
    """Tests No_Blank=False for `Distinct_Count()` of PyColumn class."""
    vals = model.Tables[testingtablename].Columns[1].Distinct_Count(No_Blank=False)
    assert isinstance(vals, int64)


@pytest.mark.parametrize("model", testing_parameters)
def test_get_sample_values(model):
    """Tests for `get_sample_values()` of PyColumn class."""
    sample_vals = model.Tables[testingtablename].Columns[1].get_sample_values()
    assert isinstance(sample_vals, pd.DataFrame) or isinstance(sample_vals, int)


@pytest.mark.parametrize("model", testing_parameters)
def test_query_every_column(model):
    """Tests `Query_All()` of PyColumns class."""
    assert isinstance(model.Tables[testingtablename].Columns.Query_All(), pd.DataFrame)



"""
`document.py` is where a specific part of pytabular start. This module can
generate pages in markdown for use in Docusaurus.
"""
import logging

from pathlib import Path

from measure import PyMeasure
from table import PyTable
from column import PyColumn
from culture import PyCulture

from .pytabular import Tabular

logger = logging.getLogger("PyTabular")


class ModelDocumenter:
    """
    The ModelDocumenter class can generate documentation based on the
    tabular object model and it will generate it suitable for docusaurus.

    TODO: Add a General Pages template with Roles and RLS Expressions
    TODO: Create a Sub Page per table for all columns, instead of one big page?
    TODO: Add Depencies per Measure with correct links.
    """

    def __init__(
        self,
        model: Tabular,
        friendly_name: str = str(),
        save_location: str = "docs",
        general_page_url: str = "1-general-information.md",
        measure_page_url: str = "2-measures.md",
        table_page_url: str = "3-tables.md",
        column_page_url: str = "4-columns.md",
        roles_page_url: str = "5-roles.md",
    ):
        self.model = model
        self.model_name = friendly_name or model.Catalog or model.Database.Name
        self.friendly_name: str = str()
        self.save_path: Path
        self.save_location: str = save_location

        # Translation information
        self.culture_include: bool = False
        self.culture_selected: str = "en-US"
        self.culture_object: PyCulture = None

        # Documentation Parts
        self.general_page: str = str()
        self.measure_page: str = str()
        self.table_page: str = str()
        self.column_page: str = str()
        self.roles_page: str = str()

        self.category_file_name: str = "_category_.yml"
        self.general_page_url: str = general_page_url
        self.measure_page_url: str = measure_page_url
        self.table_page_url: str = table_page_url
        self.column_page_url: str = column_page_url
        self.roles_page_url: str = roles_page_url

        # Generate an url friendly name for the model / folder
        self.friendly_name: str = self.set_model_friendly_name()

        # Initialize Save path so checks can be run against it.
        self.save_path = self.set_save_path()

    def create_object_reference(self, object: str, object_parent: str) -> str:
        """
        Create a Custom ID for link sections in the docs.
        :This is based on the technical names in the model,
        so not the once in the translations. This makes it
        possible to link based on dependencies.
        """
        url_reference = f"{object_parent}-{object}".replace(" ", "")
        return f"{{#{url_reference}}}"

    def generate_documentation_pages(self) -> None:
        """
        Generate Documentation for each specific part of the model.
        """
        self.measure_page = self.generate_markdown_measure_page()
        self.table_page = self.generate_markdown_table_page()
        self.column_page = self.generate_markdown_column_page()
        self.category_page = self.generate_category_file()

    def get_object_caption(self, object_name: str, object_parent: str):
        """
        Retrieves the caption of an object, based on the translations
        in the culture.
        """

        if self.culture_include:
            return self.culture_object.get_translation(
                object_name=object_name, object_parent_name=object_parent
            ).get("object_translation")

        return object_name

    def set_translations(
        self, enable_translations: bool = False, culture: str = "en-US"
    ) -> None:
        """
        Set translations to active or inactive, depending on the needs of the users.
        """

        logger.info(f"Using Translations set to > {enable_translations}")

        if enable_translations:
            try:
                self.culture_object = self.model.Cultures[culture]
                self.culture_selected = culture
                self.culture_include = enable_translations
            except IndexError:
                self.culture_include = False
                logger.warn(
                    "Culture not found, reverting back to orginal setting > False"
                )
            else:
                logger.info(f"Setting culture to {self.culture_selected}")

        else:
            self.culture_include = enable_translations

    def set_model_friendly_name(self):
        """
        Replaces the model name to a friendly string,
        so it can be used in an URL.
        """
        return (self.model_name).replace(" ", "-").replace("_", "-").lower()

    def set_save_path(self) -> Path:
        """
        Set the location of the documentation
        """
        return Path(f"{self.save_location}/{self.friendly_name}")

    def save_page(self, content: str, page_name: str, keep_file: bool = False) -> None:
        """Save the content of the documentation to a file, based on
            the class setup.
                - Save Location
                - Model Friendly Name
                - Page to be written

        Args:
            content (str): File content to write to file.
            page_name (str): Name of the file that will be used.
            keep_file (bool): The file will only be overwritten if
                the keep_file is set to False.

        Returns:
            None

        """
        target_file = self.save_path / page_name

        if keep_file and target_file.exists():
            logger.info(f"{page_name} already exists -> fill will not overwritten.")
        else:
            logger.info(f"Results are written to -> {page_name}.")

            with target_file.open("w", encoding="utf-8") as f:
                f.write(content)
                f.close()

    def save_documentation(self) -> None:
        """Generate documentation of the model, based on the meta-data
            in the model definitions. This first checks if the folder
            exists, and then starts to export the files that are needed
            for the documentatation.
                - General Information Page -> Free format page to create.
                - Measure Page -> Describes the measures in the model. (Incl. OLS?)
                - Tables Page -> Describes the tables in the model. (Incl. OLS?)
                - Columns Page -> Describes all columns in the model. (Incl. OLS?)
                - Roles Page -> Describes the roles in the model, (incl. RLS?)

        Args:
            self (Docs): Model object for documentation.

        Returns:
            None
        """
        if self.save_path.exists():
            logger.info(
                f"Path exists -> Generating documentation for {self.friendly_name}"
            )
        else:
            logger.info(
                f"Path does not exist -> Creating directory for {self.friendly_name}"
            )
            self.save_path.mkdir(parents=True, exist_ok=True)

        if self.category_page:
            self.save_page(
                content=self.category_page,
                keep_file=True,
                page_name=self.category_file_name,
            )

        # Create General information page.
        if self.general_page:
            self.save_page(
                content="General Info", keep_file=True, page_name=self.general_page_url
            )

        if self.measure_page:
            self.save_page(
                content=self.measure_page,
                keep_file=False,
                page_name=self.measure_page_url,
            )

        if self.table_page:
            self.save_page(
                content=self.table_page, keep_file=False, page_name=self.table_page_url
            )

        if self.column_page:
            self.save_page(
                content=self.column_page,
                keep_file=False,
                page_name=self.column_page_url,
            )

        if self.roles_page:
            self.save_page(
                content=self.roles_page, keep_file=False, page_name=self.roles_page_url
            )

    def create_markdown_for_measure(self, object: PyMeasure) -> str:
        """
        Create Markdown for a specific measure, that can later on be used
        for generating the whole measure page.
        """
        object_caption = (
            self.get_object_caption(
                object_name=object.Name, object_parent=object.Parent.Name
            )
            or object.Name
        )

        object_description = (object.Description or "No Description available").replace(
            "\\n", ""
        )

        return f"""
### {object_caption}
**Description**:
> {object_description}

<dl>

<dt>Display Folder</dt>
<dd>{object.DisplayFolder}</dd>

<dt>Table Name</dt>
<dd>{object.Parent.Name}</dd>

<dt>Format String</dt>
<dd>{object.FormatString}</dd>

<dt>Is Hidden</dt>
<dd>{object.IsHidden}</dd>

</dl>

```dax title="Technical: {object.Name}"
{
    object.Expression
}
```

---
"""

    def generate_markdown_measure_page(self) -> str:
        """
        Based on the measure objects it generates a page based on
        the docusaurus notation for markdown pages.
        """
        prev_display_folder = ""
        markdown_template = [
            f"""---
sidebar_position: 3
title: Measures
description: This page contains all measures for the {self.model.Name} model, including the description, format string, and other technical details.
---

# Measures for {self.model.Name}
"""
        ]

        measures = sorted(
            self.model.Measures, key=lambda x: x.DisplayFolder, reverse=False
        )

        for measure in measures:
            logger.debug(f"Creating docs for {measure.Name}")
            display_folder = measure.DisplayFolder or "Other"
            display_folder = display_folder.split("\\")[0]

            if prev_display_folder != display_folder:
                markdown_template.append(f"""## {display_folder}""")
                prev_display_folder = display_folder

            markdown_template.append(self.create_markdown_for_measure(measure))

        return "".join(markdown_template)

    def create_markdown_for_table(self, object: PyTable) -> str:
        """
        This functions returns the markdwon that can be used
        for the documentation page for Tables.

        Args:
            object: PyTable -> Based on the PyTabular Package.

        Returns:
            Markdown text: str -> Will be append to the page text.
        """
        object_caption = (
            self.get_object_caption(
                object_name=object.Name, object_parent=object.Parent.Name
            )
            or object.Name
        )

        object_description = (object.Description or "No Description available").replace(
            "\\n", ""
        )

        partition_type = ""
        partition_source = ""

        if str(object.Partitions[0].SourceType) == "Calculated":
            partition_type = "dax"
            partition_source = object.Partitions[0].Source.Expression
        elif str(object.Partitions[0].SourceType) == "M":
            partition_type = "powerquery"
            partition_source = object.Partitions[0].Source.Expression
        else:
            partition_type = "sql"
            partition_source = object.Partitions[0].Source.Query

        return f"""
### {object_caption}
**Description**:
> {object_description}

<dl>
<dt>Measures (#)</dt>
<dd>{len(object.Measures)}</dd>

<dt>Columns (#)</dt>
<dd>{len(object.Columns)}</dd>

<dt>Partitions (#)</dt>
<dd>{len(object.Partitions)}</dd>

<dt>Data Category</dt>
<dd>{object.DataCategory or "Regular Table"}</dd>

<dt>Is Hidden</dt>
<dd>{object.IsHidden}</dd>

<dt>Table Type</dt>
<dd>{object.Partitions[0].ObjectType}</dd>

<dt>Source Type</dt>
<dd>{object.Partitions[0].SourceType}</dd>
</dl>

```{partition_type} title="Table Source: {object.Name}"
{
    partition_source
}
```

---

"""

    def generate_markdown_table_page(self) -> str:
        """
        This function generates the markdown tables documentation for the tables in the Model.
        """
        markdown_template = f"""---
sidebar_position: 2
title: Tables
description: This page contains all columns with tables for {self.model.Name}, including the description, and technical details.
---

# Tables {self.model.Name}

    """

        for table in self.model.Tables:
            markdown_template += self.create_markdown_for_table(table)

        return markdown_template

    def generate_markdown_column_page(self) -> str:
        """
        This function generates the markdown for documentation about columns in the Model.
        """
        markdown_template = f"""---
sidebar_position: 4
title: Columns
description: This page contains all columns with Columns for {self.model.Name}, including the description, format string, and other technical details.
---

    """

        for table in self.model.Tables:
            markdown_template += f"""
## Columns for {table.Name}
                                """

            for column in table.Columns:
                if "RowNumber" in column.Name:
                    continue

                markdown_template += self.create_markdown_for_column(column)

        return markdown_template

    def create_markdown_for_column(self, object: PyColumn) -> str:
        """
        Generates the Markdown for a specifc column.
        If a colums is calculated, then it also shows
        the expression for that column in DAX.
        """
        object_caption = (
            self.get_object_caption(
                object_name=object.Name, object_parent=object.Parent.Name
            )
            or object.Name
        )

        object_description = (object.Description or "No Description available").replace(
            "\\n", ""
        )

        basic_info = f"""
### {object_caption} {self.create_object_reference(object=object.Name, object_parent=object.Parent.Name)}
**Description**:
> {object_description}

<dl>
<dt>Column Name</dt>
<dd>{object.Name}</dd>

<dt>Object Type</dt>
<dd>{object.ObjectType}</dd>

<dt>Type</dt>
<dd>{object.Type}</dd>

<dt>Is Available In Excel</dt>
<dd>{object.IsAvailableInMDX}</dd>

<dt>Is Hidden</dt>
<dd>{object.IsHidden}</dd>

<dt>Data Category</dt>
<dd>{object.DataCategory}</dd>

<dt>Data Type</dt>
<dd>{object.DataType}</dd>

<dt>DisplayFolder</dt>
<dd>{object.DisplayFolder}</dd>

</dl>
"""

        if str(object.Type) == "Calculated":
            basic_info += f"""
```dax title="Technical: {object.Name}"
{
    object.Expression
}
```
    """
        return (
            basic_info
            + """
---
    """
        )

    def generate_category_file(self):
        """
        Docusaurs can generate an index. The category yaml will
        make that happen.
        """
        return f"""position: 2 # float position is supported
label: '{self.model_name}'
collapsible: true # make the category collapsible
collapsed: true # keep the category open by default
link:
  type: generated-index
  title: Documentation Overview
customProps:
  description: To be added in the future.
"""



"""pytest for the table.py file. Covers the PyTable and PyTables classes.
"""
from test.config import testing_parameters, testingtablename
import pytest
import pytabular as p
from test.conftest import testing_storage


@pytest.mark.parametrize("model", testing_parameters)
def test_disconnect_for_trace(model):
    """Tests `Disconnect()` from `Tabular` class."""
    model.Disconnect()
    assert model.Server.Connected is False


@pytest.mark.parametrize("model", testing_parameters)
def test_reconnect_update(model):
    """This will test the `Reconnect()` gets called in `Update()`
    of the `Base_Trace` class.
    """
    model.Disconnect()
    model.Tables[testingtablename].Refresh()
    assert model.Server.Connected is True


@pytest.mark.parametrize("model", testing_parameters)
def test_query_monitor_start(model):
    """This will test the `Query_Monitor` trace and `Start()` it."""
    query_trace = p.Query_Monitor(model)
    query_trace.Start()
    testing_storage.query_trace = query_trace
    assert testing_storage.query_trace.Trace.IsStarted


@pytest.mark.parametrize("model", testing_parameters)
def test_query_monitor_stop(model):
    """Tests `Stop()` of `Query_Monitor` trace."""
    testing_storage.query_trace.Stop()
    assert testing_storage.query_trace.Trace.IsStarted is False


@pytest.mark.parametrize("model", testing_parameters)
def test_query_monitor_drop(model):
    """Tests `Drop()` of `Query_Monitor` trace."""
    assert testing_storage.query_trace.Drop() is None



"""
This has a `Tabular_Editor` class which will download TE2 from a default location.
Or you can input your own location.
"""
import logging
import os
import requests as r
import zipfile as Z
import atexit
from logic_utils import remove_folder_and_contents

logger = logging.getLogger("PyTabular")


def Download_Tabular_Editor(
    Download_Location: str = "https://github.com/TabularEditor/TabularEditor/releases/download/2.16.7/TabularEditor.Portable.zip",
    Folder: str = "Tabular_Editor_2",
    Auto_Remove=True,
) -> str:
    """Runs a request.get() to retrieve the zip file from web. Will unzip response and store in directory. Will also register the removal of the new directory and files when exiting program.

    Args:
            Download_Location (str, optional): File path for zip of Tabular Editor 2. Defaults to [Tabular Editor 2 Github Zip Location]'https://github.com/TabularEditor/TabularEditor/releases/download/2.16.7/TabularEditor.Portable.zip'.
            Folder (str, optional): New Folder Location. Defaults to 'Tabular_Editor_2'.
            Auto_Remove (bool, optional): Boolean to determine auto removal of files once script exits. Defaults to True.

    Returns:
            str: File path of TabularEditor.exe
    """
    logger.info("Downloading Tabular Editor 2...")
    logger.info(f"From... {Download_Location}")
    folder_location = os.path.join(os.getcwd(), Folder)
    response = r.get(Download_Location)
    file_location = f"{os.getcwd()}\\{Download_Location.split('/')[-1]}"
    with open(file_location, "wb") as te2_zip:
        te2_zip.write(response.content)
    with Z.ZipFile(file_location) as zipper:
        zipper.extractall(path=folder_location)
    logger.debug("Removing Zip File...")
    os.remove(file_location)
    logger.info(f"Tabular Editor Downloaded and Extracted to {folder_location}")
    if Auto_Remove:
        logger.debug(f"Registering removal on termination... For {folder_location}")
        atexit.register(remove_folder_and_contents, folder_location)
    return f"{folder_location}\\TabularEditor.exe"


class Tabular_Editor:
    """Setting Tabular_Editor Class for future work.
    Mainly runs `Download_Tabular_Editor()`
    """

    def __init__(self, EXE_File_Path: str = "Default") -> None:
        logger.debug(f"Initializing Tabular Editor Class:: {EXE_File_Path}")
        if EXE_File_Path == "Default":
            self.EXE: str = Download_Tabular_Editor()
        else:
            self.EXE: str = EXE_File_Path
        pass



"""`logic_utils` used to store multiple functions that are used in many different files.
"""
import logging
import datetime
import os
from typing import Dict, List
import pandas as pd
from Microsoft.AnalysisServices.Tabular import DataType
from Microsoft.AnalysisServices.AdomdClient import AdomdDataReader

logger = logging.getLogger("PyTabular")


def ticks_to_datetime(ticks: int) -> datetime.datetime:
    """Converts a C# System DateTime Tick into a Python DateTime

    Args:
            ticks (int): [C# DateTime Tick](https://docs.microsoft.com/en-us/dotnet/api/system.datetime.ticks?view=net-6.0)

    Returns:
            datetime.datetime: [datetime.datetime](https://docs.python.org/3/library/datetime.html)
    """
    return datetime.datetime(1, 1, 1) + datetime.timedelta(microseconds=ticks // 10)


def pandas_datatype_to_tabular_datatype(df: pd.DataFrame) -> Dict:
    """WiP takes dataframe columns and gets respective tabular column datatype.  ([NumPy Datatypes](https://numpy.org/doc/stable/reference/generated/numpy.dtype.kind.html) and [Tabular Datatypes](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.datatype?view=analysisservices-dotnet))

    Args:
            df (pd.DataFrame): Pandas DataFrame

    Returns:
            Dict: EX {'col1': <Microsoft.AnalysisServices.Tabular.DataType object at 0x0000023BFFBC9700>, 'col2': <Microsoft.AnalysisServices.Tabular.DataType object at 0x0000023BFFBC8840>, 'col3': <Microsoft.AnalysisServices.Tabular.DataType object at 0x0000023BFFBC9800>}
    """
    logger.info("Getting DF Column Dtypes to Tabular Dtypes...")
    tabular_datatype_mapping_key = {
        "b": DataType.Boolean,
        "i": DataType.Int64,
        "u": DataType.Int64,
        "f": DataType.Double,
        "c": DataType.Double,
        "m": DataType.DateTime,
        "M": DataType.DateTime,
        "O": DataType.String,
        "S": DataType.String,
        "U": DataType.String,
        "V": DataType.String,
    }
    return {
        column: tabular_datatype_mapping_key[df[column].dtype.kind]
        for column in df.columns
    }


def pd_dataframe_to_m_expression(df: pd.DataFrame) -> str:
    """This will take a pandas dataframe and convert to an m expression
    For example this DF:
       col1  col2
    0   1     3
    1   2     4

    |
    |
    V

    Will convert to this expression string:
    let
    Source=#table({"col1","col2"},
    {
    {"1","3"},{"2","4"}
    })
    in
    Source

    Args:
            df (pd.DataFrame): Pandas DataFrame

    Returns:
            str: Currently only returning string values in your tabular model.
    """

    def m_list_expression_generator(list_of_strings: List[str]) -> str:
        """
        Takes a python list of strings and converts to power query m expression list format...
        Ex: ["item1","item2","item3"] --> {"item1","item2","item3"}
        Codepoint reference --> \u007b == { and \u007d == }
        """
        string_components = ",".join(
            [f'"{string_value}"' for string_value in list_of_strings]
        )
        return f"\u007b{string_components}\u007d"

    logger.debug(f"Executing m_list_generator()... for {df.columns}")
    columns = m_list_expression_generator(df.columns)
    expression_str = f"let\nSource=#table({columns},\n"
    logger.debug(
        f"Iterating through rows to build expression... df has {len(df)} rows..."
    )
    expression_list_rows = []
    for _, row in df.iterrows():
        expression_list_rows += [m_list_expression_generator(row.to_list())]
    expression_str += f"\u007b\n{','.join(expression_list_rows)}\n\u007d)\nin\nSource"
    return expression_str


def remove_folder_and_contents(folder_location):
    """Internal used in tabular_editor.py and best_practice_analyzer.py.

    Args:
            folder_location (str): Folder path to remove directory and contents.
    """
    import shutil

    if os.path.exists(folder_location):
        logger.info(f"Removing Dir and Contents -> {folder_location}")
        shutil.rmtree(folder_location)


def remove_file(file_path):
    """Just `os.remove()` but wanted a `logger.info()` with it.

    Args:
        file_path: [See os.remove](https://docs.python.org/3/library/os.html)
    """
    logger.info(f"Removing file - {file_path}")
    os.remove(file_path)
    pass


def remove_suffix(input_string, suffix):
    """Adding for >3.9 compatiblity. [Stackoverflow Answer](https://stackoverflow.com/questions/66683630/removesuffix-returns-error-str-object-has-no-attribute-removesuffix)

    Args:
            input_string (str): input string to remove suffix from
            suffix (str): suffix to be removed

    Returns:
            str: input_str with suffix removed
    """
    output = (
        input_string[: -len(suffix)]
        if suffix and input_string.endswith(suffix)
        else input_string
    )
    return output


def get_sub_list(lst: list, n: int) -> list:
    """Nest list by n amount...
    `get_sub_list([1,2,3,4,5,6],2) == [[1,2],[3,4],[5,6]]`

    Args:
        lst (list): List to nest.
        n (int): Amount to nest list.

    Returns:
        list: Nested list.
    """
    return [lst[i : i + n] for i in range(0, len(lst), n)]


def get_value_to_df(Query: AdomdDataReader, index: int):
    """Gets the values from the AdomdDataReader to convert the .Net Object
    into a tangible python value to work with in pandas.
    Lots of room for improvement on this one.

    Args:
        Query (AdomdDataReader): [AdomdDataReader](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.adomdclient.adomddatareader?view=analysisservices-dotnet)
        index (int): Index of the value to perform the logic on.
    """
    if (
        Query.GetDataTypeName((index)) in ("Decimal")
        and Query.GetValue(index) is not None
    ):
        return Query.GetValue(index).ToDouble(Query.GetValue(index))
    else:
        return Query.GetValue(index)


def dataframe_to_dict(df: pd.DataFrame) -> List[dict]:
    """Convert to Dataframe to dictionary and alter columns names with;
    - Underscores (_) to spaces
    - All Strings are converted to Title Case.

    Args:
        df (pd.DataFrame): Original table that needs to be converted
            to a list with dicts.

    Returns:
        list of dictionaries.
    """
    list_of_dicts = df.to_dict("records")
    return [
        {k.replace("_", " ").title(): v for k, v in dict.items()}
        for dict in list_of_dicts
    ]


def dict_to_markdown_table(list_of_dicts: list, columns_to_include: list = None) -> str:
    """
    Description: Generate a Markdown table based on a list of dictionaries.
    Args:
        list_of_dicts (list): List of Dictionaries that need to be converted
            to a markdown table.
        columns_to_include (list): Default = None, and all colums are included.
            If a list is supplied, those columns will be included.

    Returns:
        String that will represent a table in Markdown.

    Example:
        columns = ['Referenced Object Type', 'Referenced Table', 'Referenced Object']
        dict_to_markdown_table(dependancies, columns)

    Result:
        | Referenced Object Type | Referenced Table | Referenced Object               |
        | ---------------------- | ---------------- | ------------------------------- |
        | TABLE                  | Cases            | Cases                           |
        | COLUMN                 | Cases            | IsClosed                        |
        | CALC_COLUMN            | Cases            | Resolution Time (Working Hours) |

    """
    keys = set().union(*[set(d.keys()) for d in list_of_dicts])

    if columns_to_include is not None:
        keys = list(keys.intersection(columns_to_include))

    table_header = f"| {' | '.join(map(str, keys))} |"
    table_header_separator = "|-----" * len(keys) + "|"
    markdown_table = [table_header, table_header_separator]

    for row in list_of_dicts:
        table_row = f"| {' | '.join(str(row.get(key, '')) for key in keys)} |"
        markdown_table.append(table_row)
    return "\n".join(markdown_table)



"""
`query.py` houses a custom `Connection` class that uses the .Net AdomdConnection.
`Connection` is created automatically when connecting to your model.
"""
import logging
import os
from typing import Union
from logic_utils import get_value_to_df
import pandas as pd
from Microsoft.AnalysisServices.AdomdClient import AdomdCommand, AdomdConnection


logger = logging.getLogger("PyTabular")


class Connection(AdomdConnection):
    """Subclass for [Adomdclient](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.adomdclient?view=analysisservices-dotnet). With some extra items on top.
    Right now designed for internal use. For example, Query method in the Tabular class is just a wrapper for this class' Query method... So use that instead.

    Args:
        AdomdConnection (_type_): _description_
    """

    def __init__(self, Server, Effective_User=None) -> None:
        super().__init__()
        if Server.ConnectionInfo.Password is None:
            connection_string = Server.ConnectionString
        else:
            connection_string = (
                f"{Server.ConnectionString}Password='{Server.ConnectionInfo.Password}'"
            )
        logger.debug(f"{connection_string}")
        if Effective_User is not None:
            connection_string += f";EffectiveUserName={Effective_User}"
        self.ConnectionString = connection_string

    def Query(self, Query_Str: str) -> Union[pd.DataFrame, str, int]:
        """Executes Query on Model and Returns Results in Pandas DataFrame

        Args:
                Query_Str (str): Dax Query. Note, needs full syntax (ex: EVALUATE). See (DAX Queries)[https://docs.microsoft.com/en-us/dax/dax-queries].
                Will check if query string is a file. If it is, then it will perform a query on whatever is read from the file.
                It is also possible to query DMV. For example. Query("select * from $SYSTEM.DISCOVER_TRACE_EVENT_CATEGORIES"). See (DMVs)[https://docs.microsoft.com/en-us/analysis-services/instances/use-dynamic-management-views-dmvs-to-monitor-analysis-services?view=asallproducts-allversions]

        Returns:
                pd.DataFrame: Returns dataframe with results
        """
        try:
            is_file = os.path.isfile(Query_Str)
        except Exception:
            is_file = False

        if is_file:
            logger.debug(
                f"File path detected, reading file... -> {Query_Str}",
            )
            with open(Query_Str, "r") as file:
                Query_Str = str(file.read())

        if str(self.get_State()) != "Open":
            # Works for now, need to update to handle different types of conneciton properties
            # https://learn.microsoft.com/en-us/dotnet/api/system.data.connectionstate?view=net-7.0
            logger.info("Checking initial Adomd Connection...")
            self.Open()
            logger.info(f"Connected! Session ID - {self.SessionID}")

        logger.debug("Querying Model...")
        logger.debug(Query_Str)
        Query = AdomdCommand(Query_Str, self).ExecuteReader()
        Column_Headers = [
            (index, Query.GetName(index)) for index in range(0, Query.FieldCount)
        ]
        Results = list()
        while Query.Read():
            Results.append(
                [
                    get_value_to_df(Query, index)
                    for index in range(0, len(Column_Headers))
                ]
            )

        Query.Close()
        logger.debug("Data retrieved... reading...")
        df = pd.DataFrame(Results, columns=[value for _, value in Column_Headers])
        if len(df) == 1 and len(df.columns) == 1:
            return df.iloc[0][df.columns[0]]
        return df



"""
`table.py` houses the main `PyTable` and `PyTables` class.
Once connected to your model, interacting with table(s) will be done through these classes.
"""
import logging
from object import PyObject
import pandas as pd
from partition import PyPartition, PyPartitions
from column import PyColumn, PyColumns
from measure import PyMeasure, PyMeasures
from pytabular.object import PyObjects
from logic_utils import ticks_to_datetime
from datetime import datetime

logger = logging.getLogger("PyTabular")


class PyTable(PyObject):
    """Wrapper for [Table Class](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.table?view=analysisservices-dotnet).
    With a few other bells and whistles added to it.
    Notice the `PyObject` magic method `__getattr__()` will search in `self._object`
    if it is unable to find it in the default attributes.
    This let's you also easily check the default .Net properties.
    """

    def __init__(self, object, model) -> None:
        super().__init__(object)
        self.Model = model
        self.Partitions = PyPartitions(
            [
                PyPartition(partition, self)
                for partition in self._object.Partitions.GetEnumerator()
            ]
        )
        self.Columns = PyColumns(
            [PyColumn(column, self) for column in self._object.Columns.GetEnumerator()]
        )
        self.Measures = PyMeasures(
            [
                PyMeasure(measure, self)
                for measure in self._object.Measures.GetEnumerator()
            ]
        )
        self._display.add_row("# of Partitions", str(len(self.Partitions)))
        self._display.add_row("# of Columns", str(len(self.Columns)))
        self._display.add_row(
            "# of Measures", str(len(self.Measures)), end_section=True
        )
        self._display.add_row("Description", self._object.Description, end_section=True)
        self._display.add_row("DataCategory", str(self._object.DataCategory))
        self._display.add_row("IsHidden", str(self._object.IsHidden))
        self._display.add_row("IsPrivate", str(self._object.IsPrivate))
        self._display.add_row(
            "ModifiedTime",
            ticks_to_datetime(self._object.ModifiedTime.Ticks).strftime(
                "%m/%d/%Y, %H:%M:%S"
            ),
        )

    def Row_Count(self) -> int:
        """Method to return count of rows. Simple Dax Query:
        `EVALUATE {COUNTROWS('Table Name')}`

        Returns:
            int: Number of rows using [COUNTROWS](https://learn.microsoft.com/en-us/dax/countrows-function-dax).
        """
        return self.Model.Adomd.Query(f"EVALUATE {{COUNTROWS('{self.Name}')}}")

    def Refresh(self, *args, **kwargs) -> pd.DataFrame:
        """Same method from Model Refresh, you can pass through any extra parameters. For example:
        `Tabular().Tables['Table Name'].Refresh(Tracing = True)`
        Returns:
            pd.DataFrame: Returns pandas dataframe with some refresh details
        """
        return self.Model.Refresh(self, *args, **kwargs)

    def Last_Refresh(self) -> datetime:
        """Will query each partition for the last refresh time then select the max

        Returns:
            datetime: Last refresh time in datetime format
        """
        partition_refreshes = [
            partition.Last_Refresh() for partition in self.Partitions
        ]
        return max(partition_refreshes)

    def Related(self):
        """Returns tables with a relationship with the table in question."""
        return self.Model.Relationships.Related(self)


class PyTables(PyObjects):
    """
    Groups together multiple tables. See `PyObjects` class for what more it can do.
    You can interact with `PyTables` straight from model. For ex: `model.Tables`.
    You can even filter down with `.Find()`. For example find all tables with `fact` in name.
    `model.Tables.Find('fact')`.
    """

    def __init__(self, objects) -> None:
        super().__init__(objects)

    def Refresh(self, *args, **kwargs):
        """Refreshes all `PyTable`(s) in class."""
        model = self._objects[0].Model
        return model.Refresh(self, *args, **kwargs)

    def Query_All(self, query_function: str = "COUNTROWS(_)") -> pd.DataFrame:
        """This will dynamically create a query to pull all tables from the model and run the query function.
        It will replace the _ with the table to run.

        Args:
                query_function (str, optional): Dax query is dynamically building a query with the UNION & ROW DAX Functions. Defaults to 'COUNTROWS(_)'.

        Returns:
                pd.DataFrame: Returns dataframe with results
        """
        logger.info("Querying every table in PyTables...")
        logger.debug(f"Function to be run: {query_function}")
        logger.debug("Dynamically creating DAX query...")
        query_str = "EVALUATE UNION(\n"
        for table in self:
            table_name = table.get_Name()
            dax_table_identifier = f"'{table_name}'"
            query_str += f"ROW(\"Table\",\"{table_name}\",\"{query_function}\",{query_function.replace('_',dax_table_identifier)}),\n"
        query_str = f"{query_str[:-2]})"
        return self[0].Model.Query(query_str)

    def Find_Zero_Rows(self):
        """Returns PyTables class of tables with zero rows queried."""
        query_function: str = "COUNTROWS(_)"
        df = self.Query_All(query_function)

        table_names = df[df[f"[{query_function}]"].isna()]["[Table]"].to_list()
        logger.debug(f"Found {table_names}")
        tables = [self[name] for name in table_names]
        return self.__class__(tables)

    def Last_Refresh(self, group_partition: bool = True) -> pd.DataFrame:
        """Returns pd.DataFrame of tables with their latest refresh time.
        Optional 'group_partition' variable, default is True.
        If False an extra column will be include to have the last refresh time to the grain of the partition
        Example to add to model model.Create_Table(p.Table_Last_Refresh_Times(model),'RefreshTimes')

        Args:
                model (pytabular.Tabular): Tabular Model
                group_partition (bool, optional): Whether or not you want the grain of the dataframe to be by table or by partition. Defaults to True.

        Returns:
                pd.DataFrame: pd dataframe with the RefreshedTime property: https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.partition.refreshedtime?view=analysisservices-dotnet#microsoft-analysisservices-tabular-partition-refreshedtime
                If group_partition == True and the table has multiple partitions, then df.groupby(by["tables"]).max()
        """
        data = {
            "Tables": [
                partition.Table.Name for table in self for partition in table.Partitions
            ],
            "Partitions": [
                partition.Name for table in self for partition in table.Partitions
            ],
            "RefreshedTime": [
                partition.Last_Refresh()
                for table in self
                for partition in table.Partitions
            ],
        }
        df = pd.DataFrame(data)
        if group_partition:
            logger.debug("Grouping together to grain of Table")
            return (
                df[["Tables", "RefreshedTime"]]
                .groupby(by=["Tables"])
                .max()
                .reset_index(drop=False)
            )
        else:
            logger.debug("Returning DF")
            return df



"""
`relationship.py` houses the main `PyRelationship` and `PyRelationships` class.
Once connected to your model, interacting with relationship(s) will be done through these classes.
"""
import logging
from object import PyObject, PyObjects
from table import PyTable, PyTables

from Microsoft.AnalysisServices.Tabular import (
    CrossFilteringBehavior,
    SecurityFilteringBehavior,
)

from typing import Union

logger = logging.getLogger("PyTabular")


class PyRelationship(PyObject):
    """Wrapper for [Relationship Class](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.relationship?view=analysisservices-dotnet)."""

    def __init__(self, object, model) -> None:
        super().__init__(object)
        self.Model = model
        self.CrossFilteringBehavior = CrossFilteringBehavior(
            self.CrossFilteringBehavior.value__
        ).ToString()
        self.SecurityFilteringBehavior = SecurityFilteringBehavior(
            self.SecurityFilteringBehavior.value__
        ).ToString()
        self.To_Table = self.Model.Tables[self.ToTable.Name]
        self.To_Column = self.To_Table.Columns[self.ToColumn.Name]
        self.From_Table = self.Model.Tables[self.FromTable.Name]
        self.From_Column = self.From_Table.Columns[self.FromColumn.Name]
        self._display.add_row("Is Active", str(self.IsActive))
        self._display.add_row("Cross Filtering Behavior", self.CrossFilteringBehavior)
        self._display.add_row(
            "Security Filtering Behavior", self.SecurityFilteringBehavior
        )
        self._display.add_row(
            "From", f"'{self.From_Table.Name}'[{self.From_Column.Name}]"
        )
        self._display.add_row("To", f"'{self.To_Table.Name}'[{self.To_Column.Name}]")


class PyRelationships(PyObjects):
    """
    Groups together multiple relationships. See `PyObjects` class for what more it can do.
    You can interact with `PyRelationships` straight from model. For ex: `model.Relationships`.
    """

    def __init__(self, objects) -> None:
        super().__init__(objects)

    def Related(self, object: Union[PyTable, str]) -> PyTables:
        """Finds related tables of a given table.

        Args:
            object (Union[PyTable, str]): `PyTable` or str of table name to find related tables for.

        Returns:
            PyTables: Returns `PyTables` class of the tables in question.
        """
        table_to_find = object if isinstance(object, str) else object.Name
        to_tables = [
            rel.To_Table
            for rel in self._objects
            if rel.From_Table.Name == table_to_find
        ]
        from_tables = [
            rel.From_Table
            for rel in self._objects
            if rel.To_Table.Name == table_to_find
        ]
        to_tables += from_tables
        return PyTables(to_tables)



"""Tests to cover the document.py file
"""
from test.config import testing_parameters
import pytest
import pytabular as p
import os
from pytabular import logic_utils
from test.conftest import testing_storage


@pytest.mark.parametrize("model", testing_parameters)
def test_basic_document_funcionality(model):
    try:
        docs = p.ModelDocumenter(model=model)
        docs.generate_documentation_pages()
        docs.save_documentation()
        testing_storage.documentation_class = docs
    except Exception:
        pytest.fail("Unsuccessful documentation generation..")


def test_basic_documentation_removed():
    docs_class = testing_storage.documentation_class
    remove = f"{docs_class.save_location}/{docs_class.friendly_name}"
    logic_utils.remove_folder_and_contents(remove)
    assert os.path.exists(remove) is False



"""`column.py` houses the main `PyColumn` and `PyColumns` class.
Once connected to your model, interacting with column(s) will be done through these classes.
"""
import logging
import pandas as pd
from object import PyObject, PyObjects
from Microsoft.AnalysisServices.Tabular import ColumnType

logger = logging.getLogger("PyTabular")


class PyColumn(PyObject):
    """Wrapper for [Column](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.column?view=analysisservices-dotnet).
    With a few other bells and whistles added to it.
    Notice the `PyObject` magic method `__getattr__()` will search in `self._object`
    if it is unable to find it in the default attributes.
    This let's you also easily check the default .Net properties.

    Args:
        Table: Parent Table to the Column
    """

    def __init__(self, object, table) -> None:
        super().__init__(object)
        self.Table = table
        self._display.add_row(
            "Description", str(self._object.Description), end_section=True
        )
        self._display.add_row("DataType", str(self._object.DataType))
        self._display.add_row("EncodingHint", str(self._object.EncodingHint))
        self._display.add_row("IsAvailableInMDX", str(self._object.IsAvailableInMDX))
        self._display.add_row("IsHidden", str(self._object.IsHidden))
        self._display.add_row("IsKey", str(self._object.IsKey))
        self._display.add_row("IsNullable", str(self._object.IsNullable))
        self._display.add_row("State", str(self._object.State))
        self._display.add_row("DisplayFolder", str(self._object.DisplayFolder))

    def get_dependencies(self) -> pd.DataFrame:
        """Returns the dependant columns of a measure"""
        dmv_query = f"select * from $SYSTEM.DISCOVER_CALC_DEPENDENCY where [OBJECT] = '{self.Name}' and [TABLE] = '{self.Table.Name}'"
        return self.Table.Model.Query(dmv_query)

    def get_sample_values(self, top_n: int = 3) -> pd.DataFrame:
        """Get sample values of column."""
        column_to_sample = f"'{self.Table.Name}'[{self.Name}]"
        try:
            # adding temporary try except. TOPNSKIP will not work for directquery mode.
            # Need an efficient way to identify if query is direct query or not.
            dax_query = f"""EVALUATE
                                TOPNSKIP(
                                    {top_n},
                                    0,
                                    FILTER(
                                        VALUES({column_to_sample}),
                                        NOT ISBLANK({column_to_sample}) && LEN({column_to_sample}) > 0
                                    ),
                                    1
                                )
                                ORDER BY {column_to_sample}
                        """
            return self.Table.Model.Query(dax_query)
        except Exception:  # pragma: no cover
            # Excluding exception from coverage.py for now...
            # This is really tech debt anyways and should be replaced...
            dax_query = f"""
            EVALUATE
                TOPN(
                    {top_n},
                    FILTER(
                        VALUES({column_to_sample}),
                        NOT ISBLANK({column_to_sample}) && LEN({column_to_sample}) > 0
                    )
                )
            """
            return self.Table.Model.Query(dax_query)

    def Distinct_Count(self, No_Blank=False) -> int:
        """Get [DISTINCTCOUNT](https://learn.microsoft.com/en-us/dax/distinctcount-function-dax) of Column.

        Args:
            No_Blank (bool, optional): Ability to call [DISTINCTCOUNTNOBLANK](https://learn.microsoft.com/en-us/dax/distinctcountnoblank-function-dax). Defaults to False.

        Returns:
            int: Number of Distinct Count from column. If `No_Blank == True` then will return number of Distinct Count no blanks.
        """
        func = "DISTINCTCOUNT"
        if No_Blank:
            func += "NOBLANK"
        return self.Table.Model.Adomd.Query(
            f"EVALUATE {{{func}('{self.Table.Name}'[{self.Name}])}}"
        )

    def Values(self) -> pd.DataFrame:
        """Get single column DataFrame of [VALUES](https://learn.microsoft.com/en-us/dax/values-function-dax)

        Returns:
            pd.DataFrame: Single Column DataFrame of Values.
        """
        return self.Table.Model.Adomd.Query(
            f"EVALUATE VALUES('{self.Table.Name}'[{self.Name}])"
        )


class PyColumns(PyObjects):
    """Groups together multiple columns. See `PyObjects` class for what more it can do.
    You can interact with `PyColumns` straight from model. For ex: `model.Columns`.
    Or through individual tables `model.Tables[TABLE_NAME].Columns`.
    You can even filter down with `.Find()`. For example find all columns with `Key` in name.
    `model.Columns.Find('Key')`.
    """

    def __init__(self, objects) -> None:
        super().__init__(objects)

    def Query_All(self, query_function: str = "COUNTROWS(VALUES(_))") -> pd.DataFrame:
        """This will dynamically create a query to pull all columns from the model and run the query function.
        It will replace the _ with the column to run.

        Args:
                query_function (str, optional): Dax query is dynamically building a query with the UNION & ROW DAX Functions.

        Returns:
                pd.DataFrame: Returns dataframe with results.
        """
        logger.info("Beginning execution of querying every column...")
        logger.debug(f"Function to be run: {query_function}")
        logger.debug("Dynamically creating DAX query...")
        query_str = "EVALUATE UNION(\n"
        columns = [column for column in self]
        for column in columns:
            if column.Type != ColumnType.RowNumber:
                table_name = column.Table.get_Name()
                column_name = column.get_Name()
                dax_identifier = f"'{table_name}'[{column_name}]"
                query_str += f"ROW(\"Table\",\"{table_name}\",\"Column\",\"{column_name}\",\"{query_function}\",{query_function.replace('_',dax_identifier)}),\n"
        query_str = f"{query_str[:-2]})"
        return self[0].Table.Model.Query(query_str)



import pytest
import pandas as pd
import pytabular as p
from test.config import testingtablename, testing_parameters, get_test_path


@pytest.mark.parametrize("model", testing_parameters)
def test_basic_query(model):
    int_result = model.Query("EVALUATE {1}")
    text_result = model.Query('EVALUATE {"Hello World"}')
    assert int_result == 1 and text_result == "Hello World"


datatype_queries = [
    ["this is a string", '"this is a string"'],
    [1, 1],
    [1000.78, "CONVERT(1000.78,CURRENCY)"],
]


@pytest.mark.parametrize("model", testing_parameters)
def test_datatype_query(model):
    for query in datatype_queries:
        result = model.Query(f"EVALUATE {{{query[1]}}}")
        assert result == query[0]


@pytest.mark.parametrize("model", testing_parameters)
def test_file_query(model):
    singlevaltest = get_test_path() + "\\singlevaltest.dax"
    dfvaltest = get_test_path() + "\\dfvaltest.dax"
    dfdupe = pd.DataFrame({"[Value1]": (1, 3), "[Value2]": (2, 4)})
    assert model.Query(singlevaltest) == 1 and model.Query(dfvaltest).equals(dfdupe)


@pytest.mark.parametrize("model", testing_parameters)
def test_repr_str(model):
    assert isinstance(model.__repr__(), str)


@pytest.mark.parametrize("model", testing_parameters)
def test_pytables_count(model):
    assert model.Tables[testingtablename].Row_Count() > 0


@pytest.mark.parametrize("model", testing_parameters)
def test_pytables_refresh(model):
    assert len(model.Tables[testingtablename].Refresh()) > 0


@pytest.mark.parametrize("model", testing_parameters)
def test_pypartitions_refresh(model):
    assert len(model.Tables[testingtablename].Partitions[0].Refresh()) > 0


@pytest.mark.parametrize("model", testing_parameters)
def test_pyobjects_adding(model):
    table = model.Tables.Find(testingtablename)
    table += table
    assert len(table) == 2


@pytest.mark.parametrize("model", testing_parameters)
def test_nonetype_decimal_bug(model):
    query_str = """
    EVALUATE
    {
        (1, CONVERT( 1.24, CURRENCY ), "Hello"), (2, CONVERT( 87661, CURRENCY ), "World"), (3,,"Test")
    }
    """
    assert len(model.Query(query_str)) == 3


@pytest.mark.parametrize("model", testing_parameters)
def test_Table_Last_Refresh_Times(model):
    """Really just testing the the function completes successfully and returns df"""
    assert isinstance(model.Tables.Last_Refresh(), pd.DataFrame)


@pytest.mark.parametrize("model", testing_parameters)
def test_Return_Zero_Row_Tables(model):
    """Testing that `Return_Zero_Row_Tables`"""
    assert isinstance(model.Tables.Find_Zero_Rows(), p.pytabular.PyTables)


@pytest.mark.parametrize("model", testing_parameters)
def test_get_dependencies(model):
    dependencies = model.Measures[0].get_dependencies()
    assert len(dependencies) > 0


@pytest.mark.parametrize("model", testing_parameters)
def test_disconnect(model):
    """Tests `Disconnect()` from `Tabular` class."""
    model.Disconnect()
    assert model.Server.Connected is False


@pytest.mark.parametrize("model", testing_parameters)
def test_reconnect(model):
    """Tests `Reconnect()` from `Tabular` class."""
    model.Reconnect()
    assert model.Server.Connected is True


@pytest.mark.parametrize("model", testing_parameters)
def test_reconnect_savechanges(model):
    """This will test the `Reconnect()` gets called in `SaveChanges()`"""
    model.Disconnect()
    model.SaveChanges()
    assert model.Server.Connected is True


@pytest.mark.parametrize("model", testing_parameters)
def test_is_process(model):
    """Checks that `Is_Process()` from `Tabular` class returns bool"""
    assert isinstance(model.Is_Process(), bool)


@pytest.mark.parametrize("model", testing_parameters)
def test_bad_table(model):
    """Checks for unable to find table exception"""
    with pytest.raises(Exception):
        model.Refresh("badtablename")


@pytest.mark.parametrize("model", testing_parameters)
def test_refresh_dict(model):
    """Checks for refreshing dictionary"""
    table = model.Tables[testingtablename]
    refresh = model.Refresh({table.Name: table.Partitions[0].Name})
    assert isinstance(refresh, pd.DataFrame)


@pytest.mark.parametrize("model", testing_parameters)
def test_refresh_dict_pypartition(model):
    """Checks for refreshing dictionary"""
    table = model.Tables[testingtablename]
    refresh = model.Refresh({table.Name: table.Partitions[0]})
    assert isinstance(refresh, pd.DataFrame)


@pytest.mark.parametrize("model", testing_parameters)
def test_bad_partition(model):
    """Checks for refreshing dictionary"""
    table = model.Tables[testingtablename]
    with pytest.raises(Exception):
        model.Refresh({table.Name: table.Partitions[0].Name + "fail"})



"""`culture.py` is used to house the `PyCulture`, `PyCultures`, and `PyObjectTranslations` classes.
"""
import logging
from object import PyObject, PyObjects
from typing import List

logger = logging.getLogger("PyTabular")


class PyCulture(PyObject):
    """Wrapper for [Cultures](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.culture?view=analysisservices-dotnet).

    Args:
        Table: Parent Table to the Object Translations
    """

    def __init__(self, object, model) -> None:
        super().__init__(object)
        self.Model = model
        self._display.add_row("Culture Name", self._object.Name)
        self.ObjectTranslations = self.set_translation()

    def set_translation(self) -> List[dict]:
        """
        Based on the culture, it creates a list of dicts
        with all the available translations in the file.
        """
        return [
            {
                "object_translation": translation.Value,
                "object_name": translation.Object.Name,
                "object_parent_name": translation.Object.Parent.Name,
                "object_type": str(translation.Property),
            }
            for translation in self._object.ObjectTranslations
        ]

    def get_translation(
        self, object_name: str, object_parent_name: str, object_type: str = "Caption"
    ) -> dict:
        """
        Get Translation makes it possible to seach a specific translation of an object.
        By default it will search for the "Caption" object type, due to fact that a
        Display folder and Description can also have translations.
        """
        try:
            # Removed walrus operator so it can be compatible with with python versions before 3.8
            translations = [
                d
                for d in self.ObjectTranslations
                if d["object_name"] == object_name
                and d["object_type"] == object_type
                and d["object_parent_name"] == object_parent_name
            ]
            return translations[0]
        except Exception:
            return {"object_not_found": "Not Available"}


class PyCultures(PyObjects):
    """Houses grouping of `PyCulture`."""

    def __init__(self, objects) -> None:
        super().__init__(objects)



"""pytest for the table.py file. Covers the PyTable and PyTables classes.
"""
from test.config import testing_parameters
import pytest
from pytabular import logic_utils
import pandas as pd
import os

sublists = [pytest.param(range(0, 50), x, id=f"Sublist {str(x)}") for x in range(1, 5)]


@pytest.mark.parametrize("lst, n", sublists)
def test_sub_list(lst, n):
    """Tests `get_sub_list()` from `logic_utils.py`."""
    result = logic_utils.get_sub_list(lst, n)
    assert isinstance(result, list)


suffix_list = [
    "helloworld_backup",
    "helloworld_testing",
    "helloworld_suffix",
    "nosuffix",
]


@pytest.mark.parametrize(
    "file_name, suffix",
    [
        pytest.param(
            suffix, suffix[suffix.find("_") + 1 :], id=suffix[suffix.find("_") + 1 :]
        )
        for suffix in suffix_list
    ],
)
def test_remove_suffix(file_name, suffix):
    result = logic_utils.remove_suffix(file_name, suffix)
    assert suffix not in result


dfs = [
    pytest.param(pd.DataFrame({"column_1": [0, 1, 2, 3, 4, 5]}), id="DataFrame1"),
    pytest.param(pd.DataFrame({"column_1": ["one", "two", "three"]}), id="DataFrame2"),
]


@pytest.mark.parametrize("df", dfs)
def test_dataframe_to_dict(df):
    assert isinstance(logic_utils.dataframe_to_dict(df), list)


@pytest.mark.parametrize("model", testing_parameters)
def test_dict_to_markdown_table(model):
    dependencies = [measure.get_dependencies() for measure in model.Measures]
    columns = ["Referenced Object Type", "Referenced Table", "Referenced Object"]
    result = logic_utils.dict_to_markdown_table(dependencies, columns)
    assert isinstance(result, str)


def test_remove_dir():
    dir = "testing_to_be_deleted"
    os.makedirs(dir)
    remove = f"{os.getcwd()}\\{dir}"
    logic_utils.remove_folder_and_contents(remove)
    assert dir not in os.listdir()


def test_remove_file():
    file_to_delete = "testing_to_be_deleted.txt"
    with open(file_to_delete, "w") as f:
        f.write("Delete this file...")
    logic_utils.remove_file(f"{os.getcwd()}\\{file_to_delete}")
    assert file_to_delete not in os.listdir()



"""pytest for the table.py file. Covers the PyTable and PyTables classes.
"""
from test.config import testing_parameters
import pytest
from pytabular import Tabular


@pytest.mark.parametrize("model", testing_parameters)
def test_rich_repr_model(model):
    try:
        model.__rich_repr__()
    except Exception:
        pytest.fail("__rich_repr__() failed")


@pytest.mark.parametrize("model", testing_parameters)
def test_rich_repr_table(model):
    try:
        model.Tables[0].__rich_repr__()
    except Exception:
        pytest.fail("__rich_repr__() failed")


@pytest.mark.parametrize("model", testing_parameters)
def test_rich_repr_tables(model):
    try:
        model.Tables.__rich_repr__()
    except Exception:
        pytest.fail("__rich_repr__() failed")


@pytest.mark.parametrize("model", testing_parameters)
def test_rich_repr_column(model):
    try:
        model.Columns[0].__rich_repr__()
    except Exception:
        pytest.fail("__rich_repr__() failed")


@pytest.mark.parametrize("model", testing_parameters)
def test_rich_repr_columns(model):
    try:
        model.Columns.__rich_repr__()
    except Exception:
        pytest.fail("__rich_repr__() failed")


@pytest.mark.parametrize("model", testing_parameters)
def test_rich_repr_partition(model):
    try:
        model.Partitions[0].__rich_repr__()
    except Exception:
        pytest.fail("__rich_repr__() failed")


@pytest.mark.parametrize("model", testing_parameters)
def test_rich_repr_partitions(model):
    try:
        model.Partitions.__rich_repr__()
    except Exception:
        pytest.fail("__rich_repr__() failed")


@pytest.mark.parametrize("model", testing_parameters)
def test_rich_repr_measure(model):
    try:
        model.Measures[0].__rich_repr__()
    except Exception:
        pytest.fail("__rich_repr__() failed")


@pytest.mark.parametrize("model", testing_parameters)
def test_rich_repr_measures(model):
    try:
        model.Measures.__rich_repr__()
    except Exception:
        pytest.fail("__rich_repr__() failed")


# TODO need to get this to have coverage for full getattr
@pytest.mark.parametrize("model", testing_parameters)
def test_get_attr(model):
    assert isinstance(model.Tables[0].Model, Tabular)


@pytest.mark.parametrize("model", testing_parameters)
def test_iadd_tables(model):
    a = model.Tables.Find("Sales")
    b = model.Tables.Find("Date")
    a += b
    assert len(a.Find("Date")) > 0


@pytest.mark.parametrize("model", testing_parameters)
def test_iadd_table(model):
    a = model.Tables.Find("Sales")
    b = model.Tables.Find("Date")[0]
    a += b
    assert len(a.Find("Date")) > 0



import pytest
from Microsoft.AnalysisServices.Tabular import Database
from test.config import testing_parameters


@pytest.mark.parametrize("model", testing_parameters)
def test_sanity_check(model):
    """Just in case... I might be crazy"""
    assert 1 == 1


@pytest.mark.parametrize("model", testing_parameters)
def test_connection(model):
    """
    Does a quick check to the Tabular Class
    To ensure that it can connnect
    """
    assert model.Server.Connected


@pytest.mark.parametrize("model", testing_parameters)
def test_database(model):
    assert isinstance(model.Database, Database)



"""`tabular_tracing.py` handles all tracing capabilities in your model.
It also includes some pre built traces to make life easier.
Feel free to build your own.
"""
import logging
import random
import xmltodict
from typing import List, Callable
from Microsoft.AnalysisServices.Tabular import Trace, TraceEvent, TraceEventHandler
from Microsoft.AnalysisServices import (
    TraceColumn,
    TraceEventClass,
    TraceEventSubclass,
)
import atexit

logger = logging.getLogger("PyTabular")


class Base_Trace:
    """Generates Trace to be run on Server.
    This is the base class to customize the type of Trace you are looking for.
    [Server Traces](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.server.traces?view=analysisservices-dotnet#microsoft-analysisservices-tabular-server-traces)

    Args:
            Tabular_Class (Tabular): Tabular Class.
            Trace_Events (List[TraceEvent]): List of Trace Events.
            [TraceEventClass](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.traceeventclass?view=analysisservices-dotnet)
            Trace_Event_Columns (List[TraceColumn]): List of Trace Event Columns.
            [TraceEventColumn](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tracecolumn?view=analysisservices-dotnet)
            Handler (Callable): Function to call when Trace returns response.
            Input needs to be two arguments.
            One is source (Which is currently None... Need to investigate why).
            Second is
            [TraceEventArgs](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.traceeventargs?view=analysisservices-dotnet)
    """

    def __init__(
        self,
        Tabular_Class,
        Trace_Events: List[TraceEvent],
        Trace_Event_Columns: List[TraceColumn],
        Handler: Callable,
    ) -> None:
        logger.debug("Trace Base Class initializing...")
        self.Name = "PyTabular_" + "".join(
            random.SystemRandom().choices(
                [str(x) for x in [y for y in range(0, 10)]], k=10
            )
        )
        self.ID = self.Name.replace("PyTabular_", "")
        self.Trace = Trace(self.Name, self.ID)
        logger.debug(f"Trace {self.Trace.Name} created...")
        self.Tabular_Class = Tabular_Class
        self.Event_Categories = self._Query_DMV_For_Event_Categories()

        self.Trace_Events = Trace_Events
        self.Trace_Event_Columns = Trace_Event_Columns
        self.Handler = Handler

        self.Build()
        self.Add()
        self.Update()
        atexit.register(self.Drop)

    def Build(self) -> bool:
        """Run on initialization.
        This will take the inputed arguments for the class
        and attempt to build the Trace.

        Returns:
                bool: True if successful
        """
        logger.info(f"Building Trace {self.Name}")
        TE = [TraceEvent(trace_event) for trace_event in self.Trace_Events]
        logger.debug(f"Adding Events to... {self.Trace.Name}")
        [self.Trace.get_Events().Add(te) for te in TE]

        def add_column(trace_event, trace_event_column):
            """Adds the column to trace event."""
            try:
                trace_event.Columns.Add(trace_event_column)
            except Exception:  # pragma: no cover
                logger.warning(f"{trace_event} - {trace_event_column} Skipped")
                pass

        logger.debug("Adding Trace Event Columns...")
        [
            add_column(trace_event, trace_event_column)
            for trace_event_column in self.Trace_Event_Columns
            for trace_event in TE
            if str(trace_event_column.value__)
            in self.Event_Categories[str(trace_event.EventID.value__)]
        ]

        logger.debug("Adding Handler to Trace...")
        self.Handler = TraceEventHandler(self.Handler)
        self.Trace.OnEvent += self.Handler
        return True

    def Add(self) -> int:
        """Runs on initialization. Adds built Trace to the Server.

        Returns:
                int: Return int of placement in Server.Traces.get_Item(int)
        """
        logger.info(f"Adding {self.Name} to {self.Tabular_Class.Server.Name}")
        return self.Tabular_Class.Server.Traces.Add(self.Trace)

    def Update(self) -> None:
        """Runs on initialization. Syncs with Server.

        Returns:
                None: Returns None.
                Unless unsuccessful then it will return the error from Server.
        """
        logger.info(f"Updating {self.Name} in {self.Tabular_Class.Server.Name}")
        if self.Tabular_Class.Server.Connected is False:
            self.Tabular_Class.Reconnect()

        return self.Trace.Update()

    def Start(self) -> None:
        """Call when you want to start the Trace

        Returns:
                None: Returns None.
                Unless unsuccessful then it will return the error from Server.
        """
        logger.info(f"Starting {self.Name} in {self.Tabular_Class.Server.Name}")
        return self.Trace.Start()

    def Stop(self) -> None:
        """Call when you want to stop the Trace

        Returns:
                None: Returns None.
                Unless unsuccessful then it will return the error from Server.
        """
        logger.info(f"Stopping {self.Name} in {self.Tabular_Class.Server.Name}")
        return self.Trace.Stop()

    def Drop(self) -> None:
        """Call when you want to drop the Trace

        Returns:
                None: Returns None. Unless unsuccessful,
                then it will return the error from Server.
        """
        logger.info(f"Dropping {self.Name} in {self.Tabular_Class.Server.Name}")
        atexit.unregister(self.Drop)
        return self.Trace.Drop()

    def _Query_DMV_For_Event_Categories(self):
        """Internal use.
        Called during the building process
        to locate allowed columns for event categories.
        This is done by executing a Tabular().Query()
        on the DISCOVER_EVENT_CATEGORIES table in the DMV.
        Then the function will parse the results,
        as it is xml inside of rows.

        Returns:
                _type_: _description_
        """
        Event_Categories = {}
        events = []
        logger.debug("Querying DMV for columns rules...")
        logger.debug("select * from $SYSTEM.DISCOVER_TRACE_EVENT_CATEGORIES")
        df = self.Tabular_Class.Query(
            "select * from $SYSTEM.DISCOVER_TRACE_EVENT_CATEGORIES"
        )
        for index, row in df.iterrows():
            xml_data = xmltodict.parse(row.Data)
            if type(xml_data["EVENTCATEGORY"]["EVENTLIST"]["EVENT"]) == list:
                events += [
                    event for event in xml_data["EVENTCATEGORY"]["EVENTLIST"]["EVENT"]
                ]
            else:
                events += [xml_data["EVENTCATEGORY"]["EVENTLIST"]["EVENT"]]
        for event in events:
            Event_Categories[event["ID"]] = [
                column["ID"] for column in event["EVENTCOLUMNLIST"]["EVENTCOLUMN"]
            ]
        return Event_Categories


def _refresh_handler(source, args):  # pragma: no cover
    """Default function called when `Refresh_Trace` is used.
    It will log various steps of the refresh process.
    """
    TextData = args.TextData.replace("<ccon>", "").replace("</ccon>", "")

    if (
        args.EventClass == TraceEventClass.ProgressReportCurrent
        and args.EventSubclass == TraceEventSubclass.ReadData
    ):
        logger.info(
            f"Total Rows Read {args.ProgressTotal} From Table '{args.ObjectPath.split('.')[-2]}' Partition '{args.ObjectPath.split('.')[-1]}' "
        )

    elif (
        args.EventClass == TraceEventClass.ProgressReportEnd
        and args.EventSubclass == TraceEventSubclass.ReadData
    ):
        if args.ProgressTotal == 0:
            logger.warning(
                f"{'::'.join(args.ObjectPath.split('.')[-2:])} QUERIED {args.ProgressTotal} ROWS!"
            )
        else:
            logger.info(
                f"Finished Reading {'::'.join(args.ObjectPath.split('.')[-2:])} for {args.ProgressTotal} Rows!"
            )

    elif args.EventSubclass == TraceEventSubclass.SwitchingDictionary:
        logger.warning(f"{TextData}")

    elif (
        args.EventClass == TraceEventClass.ProgressReportBegin
        and args.EventSubclass
        in [
            TraceEventSubclass.TabularSequencePoint,
            TraceEventSubclass.TabularRefresh,
            TraceEventSubclass.Process,
            TraceEventSubclass.VertiPaq,
            TraceEventSubclass.CompressSegment,
            TraceEventSubclass.TabularCommit,
            TraceEventSubclass.RelationshipBuildPrepare,
            TraceEventSubclass.AnalyzeEncodeData,
            TraceEventSubclass.ReadData,
        ]
    ):
        logger.info(f"{TextData}")

    elif (
        args.EventClass == TraceEventClass.ProgressReportEnd
        and args.EventSubclass
        in [
            TraceEventSubclass.TabularSequencePoint,
            TraceEventSubclass.TabularRefresh,
            TraceEventSubclass.Process,
            TraceEventSubclass.VertiPaq,
            TraceEventSubclass.CompressSegment,
            TraceEventSubclass.TabularCommit,
            TraceEventSubclass.RelationshipBuildPrepare,
            TraceEventSubclass.AnalyzeEncodeData,
        ]
    ):
        logger.info(f"{TextData}")

    else:
        logger.debug(f"{args.EventClass}::{args.EventSubclass}::{TextData}")


class Refresh_Trace(Base_Trace):
    """Subclass of Base_Trace. For built-in Refresh Tracing.
    Run by default when refreshing tables or partitions.

    Args:
            Base_Trace (Base_Trace): Base_Trace Class
    """

    def __init__(
        self,
        Tabular_Class,
        Trace_Events: List[TraceEvent] = [
            TraceEventClass.ProgressReportBegin,
            TraceEventClass.ProgressReportCurrent,
            TraceEventClass.ProgressReportEnd,
            TraceEventClass.ProgressReportError,
        ],
        Trace_Event_Columns: List[TraceColumn] = [
            TraceColumn.EventSubclass,
            TraceColumn.CurrentTime,
            TraceColumn.ObjectName,
            TraceColumn.ObjectPath,
            TraceColumn.DatabaseName,
            TraceColumn.SessionID,
            TraceColumn.TextData,
            TraceColumn.EventClass,
            TraceColumn.ProgressTotal,
        ],
        Handler: Callable = _refresh_handler,
    ) -> None:
        super().__init__(Tabular_Class, Trace_Events, Trace_Event_Columns, Handler)


def _query_monitor_handler(source, args):  # pragma: no cover
    """
    Default function used with the `Query_Monitor` trace.
    """
    total_secs = args.Duration / 1000
    domain_site = args.NTUserName.find("\\")
    if domain_site > 0:
        user = args.NTUserName[domain_site + 1 :]
    else:
        user = args.NTUserName
    logger.info(f"{args.EventSubclass} by {user} in {args.ApplicationName}")
    logger.info(
        f"From {args.StartTime} to {args.EndTime} for {str(total_secs)} seconds"
    )
    if args.Severity == 3:
        logger.error(f"Query failure... {str(args.Error)}")
        logger.error(f"{args.TextData}")
    logger.debug(f"{args.TextData}")


class Query_Monitor(Base_Trace):
    """Subclass of Base_Trace. For built-in Query Monitoring.
    If you want to see full query text, set logger to debug.

    Args:
            Base_Trace (Base_Trace): Base_Trace Class
    """

    def __init__(
        self,
        Tabular_Class,
        Trace_Events: List[TraceEvent] = [TraceEventClass.QueryEnd],
        Trace_Event_Columns: List[TraceColumn] = [
            TraceColumn.EventSubclass,
            TraceColumn.StartTime,
            TraceColumn.EndTime,
            TraceColumn.Duration,
            TraceColumn.Severity,
            TraceColumn.Error,
            TraceColumn.NTUserName,
            TraceColumn.DatabaseName,
            TraceColumn.ApplicationName,
            TraceColumn.TextData,
        ],
        Handler: Callable = _query_monitor_handler,
    ) -> None:
        super().__init__(Tabular_Class, Trace_Events, Trace_Event_Columns, Handler)
        logger.info("Query text lives in DEBUG, adjust logging to see query text.")



"""`pbi_helper.py` was reverse engineered from [DaxStudio PowerBiHelper.cs](https://github.com/DaxStudio/DaxStudio/blob/master/src/DaxStudio.UI/Utils/PowerBIHelper.cs).
So all credit and genius should go to DaxStudio.
I just wanted it in python...
The main function is `find_local_pbi_instances()`.
It will find any open PBIX files on your computer and spit out a connection string for you.
"""
import pytabular as p
import subprocess


def get_msmdsrv() -> list:
    """Runs powershel command to retrieve the ProcessId from Get-CimInstance where Name == 'msmdsrv.exe'.

    Returns:
        list: returns ProcessId(s) in list format to account for multiple PBIX files open at the same time.
    """
    p.logger.debug("Retrieving msmdsrv.exe(s)")

    try:
        msmdsrv = subprocess.check_output(
            [
                "powershell",
                """Get-CimInstance -ClassName Win32_Process -Property * -Filter "Name = 'msmdsrv.exe'" | Select-Object -Property ProcessId -ExpandProperty ProcessId""",
            ]
        )

        msmdsrv_id = msmdsrv.decode().strip().splitlines()
        p.logger.debug(f"ProcessId for msmdsrv.exe {msmdsrv_id}")
        return msmdsrv_id

    except subprocess.CalledProcessError as e:
        p.logger.error(
            f"command '{e.cmd}' return with error (code {e.returncode}): {e.output}"
        )
        p.logger.warn(
            "Check if powershell is availabe in the PATH environment variables."
        )
        raise RuntimeError(
            f"command '{e.cmd}' return with error (code {e.returncode}): {e.output}"
        ) from e


def get_port_number(msmdsrv: str) -> str:
    """Gets the local port number of given msmdsrv ProcessId. Via PowerShell.

    Args:
        msmdsrv (str): A ProcessId returned from `get_msmdsrv()`.

    Returns:
        str: `LocalPort` returned for specific ProcessId.
    """
    port = subprocess.check_output(
        [
            "powershell",
            f"""Get-NetTCPConnection -State Listen -OwningProcess {msmdsrv} | Select-Object -Property LocalPort -First 1 -ExpandProperty LocalPort""",
        ]
    )
    port_number = port.decode().strip()
    p.logger.debug(f"Listening port - {port_number} for msmdsrv.exe - {msmdsrv}")
    return port_number


def get_parent_id(msmdsrv: str) -> str:
    """Gets ParentProcessId via PowerShell from the msmdsrv ProcessId.

    Args:
        msmdsrv (str):  A ProcessId returned from `get_msmdsrv()`.

    Returns:
        str: Returns ParentProcessId in `str` format.
    """
    parent = subprocess.check_output(
        [
            "powershell",
            f"""Get-CimInstance -ClassName Win32_Process -Property * -Filter "ProcessId = {msmdsrv}" | Select-Object -Property ParentProcessId -ExpandProperty ParentProcessId""",
        ]
    )
    parent_id = parent.decode().strip()
    p.logger.debug(f"ProcessId - {parent_id} for parent of msmdsrv.exe - {msmdsrv}")
    return parent_id


def get_parent_title(parent_id: str) -> str:
    """Takes the ParentProcessId and gets the name of the PBIX file.

    Args:
        parent_id (str): Takes ParentProcessId which can be retrieved from `get_parent_id(msmdsrv)`

    Returns:
        str: Returns str of title of PBIX file.
    """
    pbi_title_suffixes: list = [
        " \u002D Power BI Desktop",  # Dash Punctuation - minus hyphen
        " \u2212 Power BI Desktop",  # Math Symbol - minus sign
        " \u2011 Power BI Desktop",  # Dash Punctuation - non-breaking hyphen
        " \u2013 Power BI Desktop",  # Dash Punctuation - en dash
        " \u2014 Power BI Desktop",  # Dash Punctuation - em dash
        " \u2015 Power BI Desktop",  # Dash Punctuation - horizontal bar
    ]
    title = subprocess.check_output(
        ["powershell", f"""(Get-Process -Id {parent_id}).MainWindowTitle"""]
    )
    title_name = title.decode().strip()
    for suffix in pbi_title_suffixes:
        title_name = title_name.replace(suffix, "")
    p.logger.debug(f"Title - {title_name} for {parent_id}")
    return title_name


def create_connection_str(port_number: str) -> str:
    """This takes the port number and adds to connection string. This is pretty bland right now, may improve later.

    Args:
        port_number (str): Port Number retrieved from `get_port_number(msmdsrv)`.

    Returns:
        str: _description_
    """
    connection_str = f"Data Source=localhost:{port_number}"
    p.logger.debug(f"Local Connection Str - {connection_str}")
    return connection_str


def find_local_pbi_instances() -> list:
    """The real genius is from this [Dax Studio PowerBIHelper](https://github.com/DaxStudio/DaxStudio/blob/master/src/DaxStudio.UI/Utils/PowerBIHelper.cs).
    I just wanted it in python not C#, so reverse engineered what DaxStudio did.
    It will run some powershell scripts to pull the appropriate info.
    Then will spit out a list with tuples inside.
    You can use the connection string to connect to your model with pytabular.

    Returns:
        list: Example -  `[('PBI File Name1','localhost:{port}'),('PBI File Name2','localhost:{port}')]`
    """
    instances = get_msmdsrv()
    pbi_instances = []
    for instance in instances:
        p.logger.debug(f"Building connection for {instance}")
        port = get_port_number(instance)
        parent = get_parent_id(instance)
        title = get_parent_title(parent)
        connect_str = create_connection_str(port)
        pbi_instances += [(title, connect_str)]
    return pbi_instances



"""
`pytabular.py` is where it all started. So there is a lot of old methods that will eventually be deprecated.
Main class is `Tabular()`. Use that for connecting with your models.
"""
import logging

from Microsoft.AnalysisServices.Tabular import (
    Server,
    ColumnType,
    Table,
    DataColumn,
    Partition,
    MPartitionSource,
)

from typing import List, Union
from collections import namedtuple
import pandas as pd
import os
import subprocess
import atexit
from logic_utils import (
    pd_dataframe_to_m_expression,
    pandas_datatype_to_tabular_datatype,
    remove_suffix,
    remove_file,
)

from table import PyTable, PyTables
from partition import PyPartitions
from column import PyColumns
from measure import PyMeasures
from culture import PyCultures, PyCulture
from relationship import PyRelationship, PyRelationships
from object import PyObject
from refresh import PyRefresh
from query import Connection

logger = logging.getLogger("PyTabular")


class Tabular(PyObject):
    """Tabular Class to perform operations: [Microsoft.AnalysisServices.Tabular](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular?view=analysisservices-dotnet). You can use this class as your main way to interact with your model.

    Args:
            CONNECTION_STR (str): Valid [Connection String](https://docs.microsoft.com/en-us/analysis-services/instances/connection-string-properties-analysis-services?view=asallproducts-allversions) for connecting to a Tabular Model.
    Attributes:
            Server (Server): See [Server MS Docs](https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.server?view=analysisservices-dotnet).
            Catalog (str): Name of Database. See [Catalog MS Docs](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.connectioninfo.catalog?view=analysisservices-dotnet#microsoft-analysisservices-connectioninfo-catalog).
            Model (Model): See [Model MS Docs](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.model?view=analysisservices-dotnet).
            AdomdConnection (AdomdConnection): For querying. See [AdomdConnection MS Docs](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.adomdclient.adomdconnection?view=analysisservices-dotnet). Connection made from parts of the originally provided connection string.
            Tables (PyTables[PyTable]): Wrappers for [Table MS Docs](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.table?view=analysisservices-dotnet). So you have the full capabilities of what the MS Docs offer and a few others. Like `Tabular().Tables['Table Name'].Row_Count()`. Or you can find a table via `Tabular().Tables[0]` or `Tabular().Tables['Table Name']`
            Columns (List[Column]): Easy access list of columns from model. See [Column MS Docs](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.column?view=analysisservices-dotnet).
            Partitions (List[Partition]): Easy access list of partitions from model. See [Partition MS Docs](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.partition?view=analysisservices-dotnet).
            Measures (List[Measure]): Easy access list of measures from model. See [Measure MS Docs](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.table.measures?view=analysisservices-dotnet#microsoft-analysisservices-tabular-table-measures).
    """

    def __init__(self, CONNECTION_STR: str):

        # Connecting to model...
        logger.debug("Initializing Tabular Class")
        self.Server = Server()
        self.Server.Connect(CONNECTION_STR)
        logger.info(f"Connected to Server - {self.Server.Name}")
        self.Catalog = self.Server.ConnectionInfo.Catalog
        logger.debug(f"Received Catalog - {self.Catalog}")
        try:
            self.Database = [
                database
                for database in self.Server.Databases.GetEnumerator()
                if database.Name == self.Catalog or self.Catalog is None
            ][0]
        except Exception:  # pragma: no cover
            err_msg = f"Unable to find Database... {self.Catalog}"
            logger.error(err_msg)
            raise Exception(err_msg)
        logger.info(f"Connected to Database - {self.Database.Name}")
        self.CompatibilityLevel: int = self.Database.CompatibilityLevel
        self.CompatibilityMode: int = self.Database.CompatibilityMode.value__
        self.Model = self.Database.Model
        logger.info(f"Connected to Model - {self.Model.Name}")
        self.Adomd: Connection = Connection(self.Server)
        self.Effective_Users: dict = {}
        self.PyRefresh = PyRefresh

        # Build PyObjects
        self.Reload_Model_Info()

        # Run subclass init
        super().__init__(self.Model)

        # Building rich table display for repr
        self._display.add_row(
            "EstimatedSize",
            f"{round(self.Database.EstimatedSize / 1000000000, 2)} GB",
            end_section=True,
        )
        self._display.add_row("# of Tables", str(len(self.Tables)))
        self._display.add_row("# of Partitions", str(len(self.Partitions)))
        self._display.add_row("# of Columns", str(len(self.Columns)))
        self._display.add_row(
            "# of Measures", str(len(self.Measures)), end_section=True
        )
        self._display.add_row("Database", self.Database.Name)
        self._display.add_row("Server", self.Server.Name)

        # Finished and registering disconnect
        logger.debug("Class Initialization Completed")
        logger.debug("Registering Disconnect on Termination...")
        atexit.register(self.Disconnect)

    def Reload_Model_Info(self) -> bool:
        """Runs on __init__ iterates through details, can be called after any model changes. Called in SaveChanges()

        Returns:
                bool: True if successful
        """
        self.Database.Refresh()

        self.Tables = PyTables(
            [PyTable(table, self) for table in self.Model.Tables.GetEnumerator()]
        )
        self.Relationships = PyRelationships(
            [
                PyRelationship(relationship, self)
                for relationship in self.Model.Relationships.GetEnumerator()
            ]
        )
        self.Partitions = PyPartitions(
            [partition for table in self.Tables for partition in table.Partitions]
        )
        self.Columns = PyColumns(
            [column for table in self.Tables for column in table.Columns]
        )
        self.Measures = PyMeasures(
            [measure for table in self.Tables for measure in table.Measures]
        )

        self.Cultures = PyCultures(
            [
                PyCulture(culture, self)
                for culture in self.Model.Cultures.GetEnumerator()
            ]
        )
        return True

    def Is_Process(self) -> bool:
        """Run method to check if Processing is occurring. Will query DMV $SYSTEM.DISCOVER_JOBS to see if any processing is happening.

        Returns:
                bool: True if DMV shows Process, False if not.
        """
        _jobs_df = self.Query("select * from $SYSTEM.DISCOVER_JOBS")
        return len(_jobs_df[_jobs_df["JOB_DESCRIPTION"] == "Process"]) > 0

    def Disconnect(self) -> None:
        """Disconnects from Model"""
        logger.info(f"Disconnecting from - {self.Server.Name}")
        atexit.unregister(self.Disconnect)
        return self.Server.Disconnect()

    def Reconnect(self) -> None:
        """Reconnects to Model"""
        logger.info(f"Reconnecting to {self.Server.Name}")
        return self.Server.Reconnect()

    def Refresh(self, *args, **kwargs) -> pd.DataFrame:
        """PyRefresh Class to handle refreshes of model.

        Args:
            model (Tabular): Main Tabular Class
            object (Union[str, PyTable, PyPartition, Dict[str, Any]]): Designed to handle a few different ways of selecting a refresh. Can be a string of 'Table Name' or dict of {'Table Name': 'Partition Name'} or even some combination with the actual PyTable and PyPartition classes.
            trace (Base_Trace, optional): Set to `None` if no Tracing is desired, otherwise you can use default trace or create your own. Defaults to Refresh_Trace.
            refresh_checks (Refresh_Check_Collection, optional): Add your `Refresh_Check`'s into a `Refresh_Check_Collection`. Defaults to Refresh_Check_Collection().
            default_row_count_check (bool, optional): Quick built in check will fail the refresh if post check row count is zero. Defaults to True.
            refresh_type (RefreshType, optional): Input RefreshType desired. Defaults to RefreshType.Full.

        Returns:
            pd.DataFrame
        """
        return self.PyRefresh(self, *args, **kwargs).Run()

    def SaveChanges(self):
        """Called after refreshes or any model changes.
        Currently will return a named tuple of all changes detected. However a ton of room for improvement here.
        """
        if self.Server.Connected is False:
            self.Reconnect()

        def property_changes(Property_Changes):
            """
            Returns any property changes.
            """
            Property_Change = namedtuple(
                "Property_Change",
                "New_Value Object Original_Value Property_Name Property_Type",
            )
            return [
                Property_Change(
                    change.NewValue,
                    change.Object,
                    change.OriginalValue,
                    change.PropertyName,
                    change.PropertyType,
                )
                for change in Property_Changes.GetEnumerator()
            ]

        logger.info("Executing SaveChanges()...")
        Model_Save_Results = self.Model.SaveChanges()
        if isinstance(Model_Save_Results.Impact, type(None)):
            logger.warning(f"No changes detected on save for {self.Server.Name}")
            return None
        else:
            Property_Changes = Model_Save_Results.Impact.PropertyChanges
            Added_Objects = Model_Save_Results.Impact.AddedObjects
            Added_Subtree_Roots = Model_Save_Results.Impact.AddedSubtreeRoots
            Removed_Objects = Model_Save_Results.Impact.RemovedObjects
            Removed_Subtree_Roots = Model_Save_Results.Impact.RemovedSubtreeRoots
            Xmla_Results = Model_Save_Results.XmlaResults
            Changes = namedtuple(
                "Changes",
                "Property_Changes Added_Objects Added_Subtree_Roots Removed_Objects Removed_Subtree_Roots Xmla_Results",
            )
            [
                property_changes(Property_Changes),
                Added_Objects,
                Added_Subtree_Roots,
                Removed_Objects,
                Removed_Subtree_Roots,
                Xmla_Results,
            ]
            self.Reload_Model_Info()
            return Changes(
                property_changes(Property_Changes),
                Added_Objects,
                Added_Subtree_Roots,
                Removed_Objects,
                Removed_Subtree_Roots,
                Xmla_Results,
            )

    def Backup_Table(self, table_str: str) -> bool:  # pragma: no cover
        """Will be removed. This is experimental with no written pytest for it.
        Backs up table in memory, brings with it measures, columns, hierarchies, relationships, roles, etc.
        It will add suffix '_backup' to all objects.
        Refresh is performed from source during backup.

        Args:
                table_str (str, optional): Name of Table.

        Returns:
                bool: Returns True if Successful, else will return error.
        """
        logger.info("Backup Beginning...")
        logger.debug(f"Cloning {table_str}")
        table = self.Model.Tables.Find(table_str).Clone()
        logger.info("Beginning Renames")

        def rename(items):
            """Iterates through items and requests rename."""
            for item in items:
                item.RequestRename(f"{item.Name}_backup")
                logger.debug(f"Renamed - {item.Name}")

        logger.info("Renaming Columns")
        rename(table.Columns.GetEnumerator())
        logger.info("Renaming Partitions")
        rename(table.Partitions.GetEnumerator())
        logger.info("Renaming Measures")
        rename(table.Measures.GetEnumerator())
        logger.info("Renaming Hierarchies")
        rename(table.Hierarchies.GetEnumerator())
        logger.info("Renaming Table")
        table.RequestRename(f"{table.Name}_backup")
        logger.info("Adding Table to Model as backup")
        self.Model.Tables.Add(table)
        logger.info("Finding Necessary Relationships... Cloning...")
        relationships = [
            relationship.Clone()
            for relationship in self.Model.Relationships.GetEnumerator()
            if relationship.ToTable.Name == remove_suffix(table.Name, "_backup")
            or relationship.FromTable.Name == remove_suffix(table.Name, "_backup")
        ]
        logger.info("Renaming Relationships")
        rename(relationships)
        logger.info("Switching Relationships to Clone Table & Column")
        for relationship in relationships:
            logger.debug(f"Renaming - {relationship.Name}")
            if relationship.ToTable.Name == remove_suffix(table.Name, "_backup"):
                relationship.set_ToColumn(
                    table.Columns.Find(f"{relationship.ToColumn.Name}_backup")
                )
            elif relationship.FromTable.Name == remove_suffix(table.Name, "_backup"):
                relationship.set_FromColumn(
                    table.Columns.Find(f"{relationship.FromColumn.Name}_backup")
                )
            logger.debug(f"Adding {relationship.Name} to {self.Model.Name}")
            self.Model.Relationships.Add(relationship)

        def clone_role_permissions():
            """Clones the role permissions for table."""
            logger.info("Beginning to handle roles and permissions for table...")
            logger.debug("Finding Roles...")
            roles = [
                role
                for role in self.Model.Roles.GetEnumerator()
                for tablepermission in role.TablePermissions.GetEnumerator()
                if tablepermission.Name == table_str
            ]
            for role in roles:
                logger.debug(f"Role {role.Name} matched, looking into it...")
                logger.debug("Searching for table specific permissions")
                tablepermissions = [
                    table.Clone()
                    for table in role.TablePermissions.GetEnumerator()
                    if table.Name == table_str
                ]
                for tablepermission in tablepermissions:
                    logger.debug(
                        f"{tablepermission.Name} found... switching table to clone"
                    )
                    tablepermission.set_Table(table)
                    for column in tablepermission.ColumnPermissions.GetEnumerator():
                        logger.debug(
                            f"Column - {column.Name} copying permissions to clone..."
                        )
                        column.set_Column(
                            self.Model.Tables.Find(table.Name).Columns.Find(
                                f"{column.Name}_backup"
                            )
                        )
                    logger.debug(f"Adding {tablepermission.Name} to {role.Name}")
                    role.TablePermissions.Add(tablepermission)
            return True

        clone_role_permissions()
        logger.info(f"Refreshing Clone... {table.Name}")
        self.Reload_Model_Info()
        self.Refresh(table.Name, default_row_count_check=False)
        logger.info(f"Updating Model {self.Model.Name}")
        self.SaveChanges()
        return True

    def Revert_Table(self, table_str: str) -> bool:  # pragma: no cover
        """Will be removed. This is experimental with no written pytest for it. This is used in conjunction with Backup_Table().
        It will take the 'TableName_backup' and replace with the original.
        Example scenario ->
        1. model.Backup_Table('TableName')
        2. perform any proposed changes in original 'TableName'
        3. validate changes in 'TableName'
        4. if unsuccessful run model.Revert_Table('TableName')

        Args:
                table_str (str): Name of table.

        Returns:
                bool: Returns True if Successful, else will return error.
        """
        logger.info(f"Beginning Revert for {table_str}")
        logger.debug(f"Finding original {table_str}")
        main = self.Model.Tables.Find(table_str)
        logger.debug(f"Finding backup {table_str}")
        backup = self.Model.Tables.Find(f"{table_str}_backup")
        logger.debug("Finding original relationships")
        main_relationships = [
            relationship
            for relationship in self.Model.Relationships.GetEnumerator()
            if relationship.ToTable.Name == main.Name
            or relationship.FromTable.Name == main.Name
        ]
        logger.debug("Finding backup relationships")
        backup_relationships = [
            relationship
            for relationship in self.Model.Relationships.GetEnumerator()
            if relationship.ToTable.Name == backup.Name
            or relationship.FromTable.Name == backup.Name
        ]

        def remove_role_permissions():
            """
            Removes role permissions from table.
            """
            logger.debug(
                f"Finding table and column permission in roles to remove from {table_str}"
            )
            roles = [
                role
                for role in self.Model.Roles.GetEnumerator()
                for tablepermission in role.TablePermissions.GetEnumerator()
                if tablepermission.Name == table_str
            ]
            for role in roles:
                logger.debug(f"Role {role.Name} Found")
                tablepermissions = [
                    table
                    for table in role.TablePermissions.GetEnumerator()
                    if table.Name == table_str
                ]
                for tablepermission in tablepermissions:
                    logger.debug(f"Removing {tablepermission.Name} from {role.Name}")
                    role.TablePermissions.Remove(tablepermission)

        for relationship in main_relationships:
            logger.debug("Cleaning relationships...")
            if relationship.ToTable.Name == main.Name:
                logger.debug(f"Removing {relationship.Name}")
                self.Model.Relationships.Remove(relationship)
            elif relationship.FromTable.Name == main.Name:
                logger.debug(f"Removing {relationship.Name}")
                self.Model.Relationships.Remove(relationship)
        logger.debug(f"Removing Original Table {main.Name}")
        self.Model.Tables.Remove(main)
        remove_role_permissions()

        def dename(items):
            """
            Denames all items.
            """
            for item in items:
                logger.debug(f"Removing Suffix for {item.Name}")
                item.RequestRename(remove_suffix(item.Name, "_backup"))
                logger.debug(f"Saving Changes... for {item.Name}")
                self.Model.SaveChanges()

        logger.info("Name changes for Columns...")
        dename(
            [
                column
                for column in backup.Columns.GetEnumerator()
                if column.Type != ColumnType.RowNumber
            ]
        )
        logger.info("Name changes for Partitions...")
        dename(backup.Partitions.GetEnumerator())
        logger.info("Name changes for Measures...")
        dename(backup.Measures.GetEnumerator())
        logger.info("Name changes for Hierarchies...")
        dename(backup.Hierarchies.GetEnumerator())
        logger.info("Name changes for Relationships...")
        dename(backup_relationships)
        logger.info("Name changes for Backup Table...")
        backup.RequestRename(remove_suffix(backup.Name, "_backup"))
        self.SaveChanges()
        return True

    def Query(
        self, Query_Str: str, Effective_User: str = None
    ) -> Union[pd.DataFrame, str, int]:
        """Executes Query on Model and Returns Results in Pandas DataFrame

        Args:
                Query_Str (str): Dax Query. Note, needs full syntax (ex: EVALUATE). See (DAX Queries)[https://docs.microsoft.com/en-us/dax/dax-queries].
                Will check if query string is a file. If it is, then it will perform a query on whatever is read from the file.
                It is also possible to query DMV. For example. Query("select * from $SYSTEM.DISCOVER_TRACE_EVENT_CATEGORIES"). See (DMVs)[https://docs.microsoft.com/en-us/analysis-services/instances/use-dynamic-management-views-dmvs-to-monitor-analysis-services?view=asallproducts-allversions]
                Effective_User (str): User you wish to query as.

        Returns:
                pd.DataFrame: Returns dataframe with results
        """
        if Effective_User is None:
            return self.Adomd.Query(Query_Str)

        try:  # pragma: no cover
            # This needs a public model with effective users to properly test
            conn = self.Effective_Users[Effective_User]
            logger.debug(f"Effective user found querying as... {Effective_User}")
        except Exception:
            logger.debug(f"Creating new connection with {Effective_User}")
            conn = Connection(self.Server, Effective_User=Effective_User)
            self.Effective_Users[Effective_User] = conn

        return conn.Query(Query_Str)

    def Query_Every_Column(
        self, query_function: str = "COUNTROWS(VALUES(_))"
    ) -> pd.DataFrame:
        """Will be removed. Use `Query_All()` in your `PyColumns` instead. This will dynamically create a query to pull all columns from the model and run the query function. It will replace the _ with the column to run.
        Args:
                query_function (str, optional): Dax query is dynamically building a query with the UNION & ROW DAX Functions.
        Returns:
                pd.DataFrame: Returns dataframe with results.
        """
        logger.info("Beginning execution of querying every column...")
        logger.warning(
            "Query_Every_Column will be deprecated... Use Query_All in PyTables class instead!"
        )
        logger.debug(f"Function to be run: {query_function}")
        logger.debug("Dynamically creating DAX query...")
        query_str = "EVALUATE UNION(\n"
        columns = [column for table in self.Tables for column in table.Columns]
        for column in columns:
            if column.Type != ColumnType.RowNumber:
                table_name = column.Table.get_Name()
                column_name = column.get_Name()
                dax_identifier = f"'{table_name}'[{column_name}]"
                query_str += f"ROW(\"Table\",\"{table_name}\",\"Column\",\"{column_name}\",\"{query_function}\",{query_function.replace('_',dax_identifier)}),\n"
        query_str = f"{query_str[:-2]})"
        return self.Query(query_str)

    def Query_Every_Table(self, query_function: str = "COUNTROWS(_)") -> pd.DataFrame:
        """Will be removed. Use `Query_All()` in your `PyTables` instead. This will dynamically create a query to pull all tables from the model and run the query function.
        It will replace the _ with the table to run.
        Args:
                query_function (str, optional): Dax query is dynamically building a query with the UNION & ROW DAX Functions. Defaults to 'COUNTROWS(_)'.
        Returns:
                pd.DataFrame: Returns dataframe with results
        """
        logger.warning(
            "Query_Every_Table will be deprecated... Use Query_All in PyTables class instead!"
        )
        logger.info("Beginning execution of querying every table...")
        logger.debug(f"Function to be run: {query_function}")
        logger.debug("Dynamically creating DAX query...")
        query_str = "EVALUATE UNION(\n"
        for table in self.Tables:
            table_name = table.get_Name()
            dax_table_identifier = f"'{table_name}'"
            query_str += f"ROW(\"Table\",\"{table_name}\",\"{query_function}\",{query_function.replace('_',dax_table_identifier)}),\n"
        query_str = f"{query_str[:-2]})"
        return self.Query(query_str)

    def Analyze_BPA(
        self, Tabular_Editor_Exe: str, Best_Practice_Analyzer: str
    ) -> List[str]:
        """Takes your Tabular Model and performs TE2s BPA. Runs through Command line.
        [Tabular Editor BPA](https://docs.tabulareditor.com/te2/Best-Practice-Analyzer.html)
        [Tabular Editor Command Line Options](https://docs.tabulareditor.com/te2/Command-line-Options.html)

        Args:
                Tabular_Editor_Exe (str): TE2 Exe File path. Feel free to use class TE2().EXE_Path or provide your own.
                Best_Practice_Analyzer (str): BPA json file path. Feel free to use class BPA().Location or provide your own. Defualts to 	https://raw.githubusercontent.com/microsoft/Analysis-Services/master/BestPracticeRules/BPARules.json

        Returns:
                List[str]: Assuming no failure, will return list of BPA violations. Else will return error from command line.
        """
        logger.debug("Beginning request to talk with TE2 & Find BPA...")
        bim_file_location = f"{os.getcwd()}\\Model.bim"
        atexit.register(remove_file, bim_file_location)
        cmd = f'{Tabular_Editor_Exe} "Provider=MSOLAP;{self.Adomd.ConnectionString}" {self.Database.Name} -B "{bim_file_location}" -A {Best_Practice_Analyzer} -V/?'
        logger.debug("Command Generated")
        logger.debug("Submitting Command...")
        sp = subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        raw_output, error = sp.communicate()
        if len(error) > 0:
            return error
        else:
            return [
                output for output in raw_output.split("\n") if "violates rule" in output
            ]

    def Create_Table(self, df: pd.DataFrame, table_name: str) -> bool:
        """Creates tables from pd.DataFrame as an M-Partition.
        So will convert the dataframe to M-Partition logic via the M query table constructor.
        Runs refresh and will update model.

        Args:
                df (pd.DataFrame): DataFrame to add to model
                table_name (str): _description_

        Returns:
                bool: True if successful
        """
        logger.debug(f"Beginning to create table for {table_name}...")
        new_table = Table()
        new_table.RequestRename(table_name)
        logger.debug("Sorting through columns...")
        df_column_names = df.columns
        dtype_conversion = pandas_datatype_to_tabular_datatype(df)
        for df_column_name in df_column_names:
            logger.debug(f"Adding {df_column_name} to Table...")
            column = DataColumn()
            column.RequestRename(df_column_name)
            column.set_SourceColumn(df_column_name)
            column.set_DataType(dtype_conversion[df_column_name])
            new_table.Columns.Add(column)
        logger.debug("Expression String Created...")
        logger.debug("Creating MPartition...")
        partition = Partition()
        partition.set_Source(MPartitionSource())
        logger.debug("Setting MPartition Expression...")
        partition.Source.set_Expression(pd_dataframe_to_m_expression(df))
        logger.debug(
            f"Adding partition: {partition.Name} to {self.Server.Name}::{self.Database.Name}::{self.Model.Name}"
        )
        new_table.Partitions.Add(partition)
        logger.debug(
            f"Adding table: {new_table.Name} to {self.Server.Name}::{self.Database.Name}::{self.Model.Name}"
        )
        self.Model.Tables.Add(new_table)
        self.SaveChanges()
        self.Reload_Model_Info()
        self.Refresh(new_table.Name)
        return True



"""
`measure.py` houses the main `PyMeasure` and `PyMeasures` class.
Once connected to your model, interacting with measure(s) will be done through these classes.
"""
import logging
import pandas as pd
from object import PyObject, PyObjects

logger = logging.getLogger("PyTabular")


class PyMeasure(PyObject):
    """Wrapper for [Microsoft.AnalysisServices.Measure](https://learn.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.measure?view=analysisservices-dotnet).
    With a few other bells and whistles added to it. WIP

    Args:
        Table: Parent Table to the Measure
    """

    def __init__(self, object, table) -> None:
        super().__init__(object)

        self.Table = table
        self._display.add_row("Expression", self._object.Expression, end_section=True)
        self._display.add_row("DisplayFolder", self._object.DisplayFolder)
        self._display.add_row("IsHidden", str(self._object.IsHidden))
        self._display.add_row("FormatString", self._object.FormatString)

    def get_dependencies(self) -> pd.DataFrame:
        """Returns the dependant objects of a measure.
        Args:
            self: The Measure Object

        Returns:
            pd.DataFrame: The Return Value is a Pandas dataframe
                            which displays all the dependancies
                            of the object.

        """
        dmv_query = f"select * from $SYSTEM.DISCOVER_CALC_DEPENDENCY where [OBJECT] = '{self.Name}' and [TABLE] = '{self.Table.Name}'"
        return self.Table.Model.Query(dmv_query)


class PyMeasures(PyObjects):
    """
    Groups together multiple measures. See `PyObjects` class for what more it can do.
    You can interact with `PyMeasures` straight from model. For ex: `model.Measures`.
    Or through individual tables `model.Tables[TABLE_NAME].Measures`.
    You can even filter down with `.Find()`. For example find all measures with `ratio` in name.
    `model.Measures.Find('ratio')`.
    """

    def __init__(self, objects) -> None:
        super().__init__(objects)



"""
`object.py` stores the main parent classes `PyObject` and `PyObjects`.
These classes are used with the others (Tables, Columns, Measures, Partitions, etc.).
"""
from abc import ABC
from rich.console import Console
from rich.table import Table
from collections.abc import Iterable


class PyObject(ABC):
    """The main parent class for your (Tables, Columns, Measures, Partitions, etc.).
    Notice the magic methods. `__rich_repr__()` starts the baseline for displaying your model.
    It uses the amazing `rich` python package and
    builds your display from the `self._display`.
    `__getattr__()` will check in `self._object`, if unable to find anything in `self`.
    This will let you pull properties from the main .Net class.
    """

    def __init__(self, object) -> None:
        self._object = object
        self._display = Table(title=self.Name)
        self._display.add_column(
            "Properties", justify="right", style="cyan", no_wrap=True
        )
        self._display.add_column("", justify="left", style="magenta", no_wrap=False)

        self._display.add_row("Name", self.Name)
        self._display.add_row("ObjectType", str(self.ObjectType))
        if str(self.ObjectType) not in "Model":
            self._display.add_row("ParentName", self.Parent.Name)
            self._display.add_row(
                "ParentObjectType",
                str(self.Parent.ObjectType),
                end_section=True,
            )

    def __rich_repr__(self) -> str:
        """See [Rich Repr Protocol](https://rich.readthedocs.io/en/stable/pretty.html#rich-repr-protocol)"""
        Console().print(self._display)

    def __getattr__(self, attr):
        """Searches in `self.__dict__` first, then `self._object`."""
        if attr in self.__dict__:
            return getattr(self, attr)
        else:
            return getattr(self._object, attr)


class PyObjects:
    """
    The main parent class for grouping your (Tables, Columns, Measures, Partitions, etc.).
    Notice the magic methods. `__rich_repr__()` starts the baseline for displaying your model.
    It uses the amazing `rich` python package and
    builds your display from the `self._display`.
    Still building out the magic methods to give `PyObjects` more flexibility.
    """

    def __init__(self, objects) -> None:
        self._objects = objects
        self._display = Table(title=str(self.__class__.mro()[0]))
        for index, obj in enumerate(self._objects):
            self._display.add_row(str(index), obj.Name)

    def __rich_repr__(self) -> str:
        """See [Rich Repr Protocol](https://rich.readthedocs.io/en/stable/pretty.html#rich-repr-protocol)"""
        Console().print(self._display)

    def __getitem__(self, object):
        """Checks if item is str or int. If string will iterate through and try to find matching name.
        Otherwise, will call into `self._objects[int]` to retrieve item.
        """
        if isinstance(object, str):
            return [pyobject for pyobject in self._objects if object == pyobject.Name][
                -1
            ]
        else:
            return self._objects[object]

    def __iter__(self):
        """Default `__iter__()` to iterate through `PyObjects`."""
        yield from self._objects

    def __len__(self) -> int:
        """Default `__len__()`

        Returns:
            int: Number of PyObject in PyObjects
        """
        return len(self._objects)

    def __iadd__(self, obj):
        """
        Add a `PyObject` or `PyObjects` to your current `PyObjects` class.
        This is useful for building out a custom `PyObjects` class to work with.
        """
        if isinstance(obj, Iterable):
            self._objects.__iadd__(obj._objects)
        else:
            self._objects.__iadd__([obj])

        self.__init__(self._objects)
        return self

    def Find(self, object_str: str):
        """Finds any or all `PyObject` inside of `PyObjects` that match the `object_str`.
        It is case insensitive.

        Args:
            object_str (str): str to lookup in `PyObjects`

        Returns:
            PyObjects: Returns a `PyObjects` class with all `PyObject` where the `PyObject.Name` matches `object_str`.
        """
        items = [
            object
            for object in self._objects
            if object_str.lower() in object.Name.lower()
        ]
        return self.__class__.mro()[0](items)



"""
All basic checks will eventually be deprecated for more intuitive methods in the right classes.
For example, instead of calling the function `Return_Zero_Row_Tables(model)`,
call the Find_Zero_Rows() method in the PyTables class.
That way you can dynamically run those methods on a subset of tables,
instead of only on the entire model.
"""
import logging
from typing import List, Union
import pytabular
from logic_utils import ticks_to_datetime
import sys
import pandas as pd
from measure import PyMeasure

logger = logging.getLogger("PyTabular")


def Return_Zero_Row_Tables(model: pytabular.Tabular) -> List[str]:
    """Returns list of table names of those that are returning isna()
    Args:
            model (pytabular.Tabular): Tabular Model
    Returns:
            List[str]: List of table names where DAX COUNTROWS('Table Name') is nan or 0.
    """
    logger.warning(
        "Return_Zero_Row_Tables() will be deprecated. \
        Instead use Zero_Row_Tables() through the PyTables class."
    )
    query_function: str = "COUNTROWS(_)"
    df: pd.DataFrame = model.Tables.Query_All(query_function)
    return df[df[f"[{query_function}]"].isna()]["[Table]"].to_list()


def Table_Last_Refresh_Times(
    model: pytabular.Tabular, group_partition: bool = True
) -> pd.DataFrame:
    """Returns pd.DataFrame of tables with their latest refresh time.
    Optional 'group_partition' variable, default is True.
    If False an extra column will be include to have the last refresh time to the grain of the partition
    Example to add to model model.Create_Table(p.Table_Last_Refresh_Times(model),'RefreshTimes')
    Args:
            model (pytabular.Tabular): Tabular Model
            group_partition (bool, optional): Whether or not you want the grain of the dataframe to be by table or by partition. Defaults to True.
    Returns:
            pd.DataFrame: pd dataframe with the RefreshedTime property: https://docs.microsoft.com/en-us/dotnet/api/microsoft.analysisservices.tabular.partition.refreshedtime?view=analysisservices-dotnet#microsoft-analysisservices-tabular-partition-refreshedtime
            If group_partition == True and the table has multiple partitions, then df.groupby(by["tables"]).max()
    """
    logger.info(f"Executing Basic Function {sys._getframe(0).f_code.co_name}")
    logger.warning(
        f"{sys._getframe(0).f_code.co_name} will be deprecated... Use Last_Refresh in PyTables class instead!"
    )

    data = {
        "Tables": [partition.Table.Name for partition in model.Partitions],
        "Partitions": [partition.Name for partition in model.Partitions],
        "RefreshedTime": [
            ticks_to_datetime(partition.RefreshedTime.Ticks).strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ"
            )[:-3]
            for partition in model.Partitions
        ],
    }
    df = pd.DataFrame(data)
    if group_partition:
        logger.debug("Grouping together to grain of Table")
        return (
            df[["Tables", "RefreshedTime"]]
            .groupby(by=["Tables"])
            .max()
            .reset_index(drop=False)
        )
    else:
        logger.debug("Returning DF")
        return df


def BPA_Violations_To_DF(model: pytabular.Tabular, te2: str, bpa: str) -> pd.DataFrame:
    """Runs BPA Analyzer from TE2 and outputs result into a DF.

    Args:
            model (pytabular.Tabular): Tabular Model Class
            te2 (str): TE2 Exe File Path (Can use TE2().EXE_path)
            bpa (str): BPA File Location (Can use BPA().Location)

    Returns:
            pd.DataFrame: Super simple right now. Just splits into two columns.. The object in violation and the rule.
    """
    logger.warning(f"{sys._getframe(0).f_code.co_name} will be deprecated...")
    logger.warning(
        "Honestly it was dumb I ever made this. You can use model.Analyze_BPA()"
    )
    results = model.Analyze_BPA(te2, bpa)
    data = [
        rule.replace(" violates rule ", "^").replace('"', "").split("^")
        for rule in results
    ]
    columns = ["Object", "Violation"]
    return pd.DataFrame(data, columns=columns)


def Last_X_Interval(
    Model: pytabular.Tabular,
    Measure: Union[str, PyMeasure],
    Column_Name: Union[str, None] = None,
    Date_Column_Identifier: str = "'Date'[DATE_DTE_KEY]",
    Number_Of_Intervals: int = 90,
    Interval: str = "DAY",
) -> pd.DataFrame:
    """Pulls the Last X Interval (Ex Last 90 Days) of a specific measure.

    Args:
            Model (pytabular.Tabular): Tabular Model to perform query on.
            Measure (Union[str,pytabular.pytabular.Measure]): Measure to query. If string, will first check for a measure in the model with that name, otherwise will assume it is a DAX Expression (Ex SUM(FactTable[ColumnValue]) ) and perform that as expression
            Column_Name (Union[str,None], optional): Column Name to be outputted in DataFrame. You can provide your own otherwise will take from the Measure Name. Defaults to "Result".
            Date_Column_Identifier (str, optional): Date column dax identifier. Defaults to "'Date'[DATE_DTE_KEY]".
            Number_Of_Intervals (int, optional): This is used to plug in the variables for [DATESINPERIOD](https://docs.microsoft.com/en-us/dax/datesinperiod-function-dax). Defaults to 90.
            Interval (str, optional): Sames as Number_Of_Intervals. Used to plug in parameters of DAX function [DATESINPERIOD](https://docs.microsoft.com/en-us/dax/datesinperiod-function-dax). Defaults to "DAY". Possible options are "DAY", "MONTH", "QUARTER", and "YEAR"

    Returns:
            pd.DataFrame: Pandas DataFrame of results.
    """
    # pragma: no cover
    logger.warning(f"{sys._getframe(0).f_code.co_name} will be deprecated...")
    logger.warning(
        "Honestly it was dumb I ever made this. Just write the darn thing in DAX"
    )
    if isinstance(PyMeasure, str):
        try:
            Measure = [
                measure for measure in Model.Measures if measure.Name == Measure
            ][-1]
            Column_Name = Measure.Name if Column_Name is None else Column_Name
            Expression = f"[{Measure.Name}]"
        except Exception:
            logging.debug("Measure is string but unable to find Measure...")
            Column_Name = "Result" if Column_Name is None else Column_Name
            Expression = Measure
    else:
        Column_Name = Measure.Name if Column_Name is None else Column_Name
        Expression = f"[{Column_Name}]"
    Query_Str = f"""
    EVALUATE
        SUMMARIZECOLUMNS(
        {Date_Column_Identifier},
        KEEPFILTERS( DATESINPERIOD ( {Date_Column_Identifier}, UTCTODAY(), -{Number_Of_Intervals}, {Interval} ) ),
        "{Column_Name}", {Expression}
        )
    """
    logging.info(
        f"Running query for {Column_Name} in the last {Number_Of_Intervals} {Interval}s..."
    )
    return Model.Query(Query_Str)



"""These are tests I have for pretty janky features of PyTabular.
These were designed selfishly for my own uses.
So seperating out, to one day sunset and remove.
"""
from test.config import testing_parameters, testingtablename
import pytest
import pytabular as p
import pandas as pd


@pytest.mark.parametrize("model", testing_parameters)
def test_backingup_table(model):
    model.Backup_Table(testingtablename)
    assert (
        len(
            [
                table
                for table in model.Model.Tables.GetEnumerator()
                if f"{testingtablename}_backup" == table.Name
            ]
        )
        == 1
    )


@pytest.mark.parametrize("model", testing_parameters)
def test_revert_table2(model):
    model.Revert_Table(testingtablename)
    assert (
        len(
            [
                table
                for table in model.Model.Tables.GetEnumerator()
                if f"{testingtablename}" == table.Name
            ]
        )
        == 1
    )


@pytest.mark.parametrize("model", testing_parameters)
def test_query_every_table_deprecate(model):
    assert len(model.Query_Every_Table()) > 0


@pytest.mark.parametrize("model", testing_parameters)
def test_query_every_column_deprecate(model):
    assert len(model.Query_Every_Column()) > 0


@pytest.mark.parametrize("model", testing_parameters)
def test_zero_row_tables_deprecate(model):
    assert isinstance(p.Return_Zero_Row_Tables(model), list)


@pytest.mark.parametrize("model", testing_parameters)
def test_last_refresh_times_deprecate(model):
    assert isinstance(p.Table_Last_Refresh_Times(model), pd.DataFrame)
